"""
$Id: struct_builder.py 2452 2020-04-23 19:20:47Z pe $

Module for building dynamic structures, e.g. configuration structure which
is customized for dynamic run time conditions.

Struct builder -- three operations:
===================================

Condition:
----------
A dictionary can be invalidated if a condition is not met. Conditions always
apply to a dictionary.
Condition object is a dictionary key that refers to either an IsIn or NotIn
object.

E.g.:
{
    'whatever': 'something',
    struct_builder.Condition('condition'): True
}
Result is {'whatever': 'something'} if 'condition' is True
Result is None in any other case.


Merge:
------
Merges a dictionary into a parent dictionary. Merge always occurs in a
dictionary (parent dictionary). The Merge Object is a dictionary key which
refers to a dictionary of cases. Each case refers to a merge dictionary (which)
should be merged into the parent dictionary.
a) If the condition is not listed in the cases, The result will be None.
b) If the merge dictionary is None (or results to None because of a nested
Condition or Merge), the parent dictionary will also result to None!
E.g.:
{
    struct_builder.Merge('condition'): {
        True: {
            'whatever': 'something',
            },
        # Default: None  # same result! (special rule a)
    }
}
Result is {'whatever': 'something'} if 'condition' is True
Result is None in any other case. (Special rules a and b)

Substitute:
-----------
Object will be substituted with a given value according to conditions.
Substitute can occur as a value (i.e. scalar, list element or dictionary value).
a) If the condition is not listed in the cases, The result will be None.
E.g.
[
    1,
    2,
    3,
    struct_builder.Substitute('bla', {
        22: 4,
        33 : 5,
        # Default: None  # same result! (special rule a)
        })
    }
]
Results to [1, 2, 3, 4] if 'bla' is 22
Results to [1, 2, 3, 5] if 'bla' is 33
Results to [1, 2, 3, None] in any other case


More complex conditions: IsIn, NotIn, BitAnd:
---------------------------------------------

For all Operations (Condition, Merge and Substitute) there is also the
possibility to have more complex conditions.
Instead of a value which is compared, IsIn, NotIn and BitAnd objects can be
used, e.g.:
{
    struct_builder.Merge('condition'): {
        1: {
            'whatever': 'something',
           },
        IsIn([2,3]): {
            'whatever': 'something else',
           },
        NotIn([7, 8, 9]): {
            'whatever': 'yet another thing',
           },
        BitAnd(2): {
            'whatever': 'the thing for 2, 3, 6, 7, ...)',
           },
        # Default: None  # same result! (special rule a)
    }
}
Note: As there is a possibility that more then one case is true, the following
priority is used:
1) equal match (case of 1: above)
2) BitAnd matches
3) IsIn matches
4) NotIn matches
5) Default

Example for Condition:
{
    'whatever': 'something',
    struct_builder.Condition('condition'): struct_builder.IsIn([True])
}
Example for Substitute:
[
    1,
    2,
    3,
    struct_builder.Substitute('bla', {
        3: 10,
        struct_builder.IsIn([1, 2, 3]): 11,
        struct_builder.NotIn([4, 5, 6]): 12,
        struct_builder.BitAnd(2): 13,
        struct_builder.Default: 14
        })
]

Special case for list:
----------------------
If a list element was a dictionary, but results to None (due to a Condition or
a nested Merge), the list element is dropped.
E.g.
[
    { 1: 2, 2: 3},
    { 1: 2, 2: 3},
    {
        1: 2, 2: 3,
        Condition('bla'): IsIn[22]
    }
]
Results in [{ 1: 2, 2: 3},{ 1: 2, 2: 3}, { 1: 2, 2: 3}] if 'bla' is 22
Results in [{ 1: 2, 2: 3},{ 1: 2, 2: 3}] in any other case
"""



class Default(object):  # pylint: disable=R0903
    # R0903: Too few public methods
    """
    Simple enumeration value for default behaviour in EbasMetadata.
    """
    pass

class IsIn(tuple):
    """
    List for validation checks.
    """
    pass

class NotIn(tuple):
    """
    Negated list for validation checks.
    """
    pass

class BitAnd(int):
    """
    Operand for bitwise "and" operation in validation checks.
    """
    pass

class Condition(object):  # pylint: disable=R0903
    # R0903: Too few public methods
    """
    Class for implementing the Condition.
    """
    def __init__(self, name):
        """
        Set up the substitution object.
        Parameters:
            name      name for the condition (str)
        Returns:
            None
        """
        self.name = name

    def evaluate(self, in_not_in, conditionals):  # pylint: disable=R0911
    # R0911: Too many return statements
        """
        Evaluates the substitution.
        Parameters:
            in_not_in       a In or NotIn Object to be used for the evaluation
            conditionals    the conditional dictionary from the StructurBuilder
        Returns:
            True/False
        """
        if self.name not in conditionals:
            raise ValueError("condition name {} not found in conditional {}"
                             .format(self.name, conditionals))
        if isinstance(in_not_in, IsIn):
            if conditionals[self.name] in in_not_in:
                return True
            else:
                return False
        if isinstance(in_not_in, NotIn):
            if conditionals[self.name] not in in_not_in:
                return True
            else:
                return False
        if isinstance(in_not_in, BitAnd):
            if conditionals[self.name] & in_not_in:
                return True
            else:
                return False
        if in_not_in == conditionals[self.name]:
            return True
        return False

class Merge(object):  # pylint: disable=R0903
    # R0903: Too few public methods
    """
    Class for implementing a Dictionary Merge.
    """
    def __init__(self, name):
        """
        Set up the substitution object.
        Parameters:
            name      name for the condition (str)
        Returns:
            None
        """
        self.name = name

    def evaluate(self, cases, conditionals):
        """
        Evaluates the substitution.
        Parameters:
            cases           the different cases and their result values (dict)
            conditionals    the conditional dictionary from the StructurBuilder
        Returns:
            dictionary which should be merged
        """
        if self.name not in conditionals:
            raise ValueError("condition name {} not found in conditional {}"
                             .format(self.name, conditionals))
        # first prio: equal value
        if conditionals[self.name] in cases:
            return cases[conditionals[self.name]]
        # second prio: Matching BitAnd exists:
        for cas in cases:
            try:
                if isinstance(cas, BitAnd) and conditionals[self.name] & cas:
                    return cases[cas]
            except TypeError:
                # ignore TypeError, there was something that cant be used for &
                # conditionals[self.name]
                pass
        # third prio: Matching IsIn exists
        for cas in cases:
            if isinstance(cas, IsIn) and  conditionals[self.name] in cas:
                return cases[cas]
        # fourth prio: Matching NotIn exists
        for cas in cases:
            if isinstance(cas, NotIn) and  conditionals[self.name] not in cas:
                return cases[cas]
        # fifth prio: Default
        if Default in cases:
            return cases[Default]
        # else None
        return None


class Substitute(object):  # pylint: disable=R0903
    # R0903: Too few public methods
    """
    Class for implementing the Substitution.
    """
    def __init__(self, name, cases):
        """
        Set up the substitution object.
        Parameters:
            name      name for the condition (str)
            cases          the different cases and their result values (dict)
        Returns:
            None
        """
        self.name = name
        self.cases = cases

    def evaluate(self, conditionals):
        """
        Evaluates the substitution.
        Parameters:
            conditionals    the conditional dictionary from the StructurBuilder
        Returns:
            Value (or struct) to be used
        """
        if self.name not in conditionals:
            raise ValueError("condition name {} not found in conditional {}"
                             .format(self.name, conditionals))
        # first prio: equal value
        if conditionals[self.name] in self.cases:
            return self.cases[conditionals[self.name]]
        # second prio: Matching BitAnd exists:
        for cas in self.cases:
            if isinstance(cas, BitAnd) and conditionals[self.name] & cas:
                return self.cases[cas]
        # third prio: Matching IsIn exists
        for cas in self.cases:
            if isinstance(cas, IsIn) and  conditionals[self.name] in cas:
                return self.cases[cas]
        # fourth prio: Matching NotIn exists
        for cas in self.cases:
            if isinstance(cas, NotIn) and  conditionals[self.name] not in cas:
                return self.cases[cas]
        # fifth prio: Default
        if Default in self.cases:
            return self.cases[Default]
        # else None
        return None


class StructureBuilder(object):  # pylint: disable=R0903
    # R0903: Too few public methods
    """
    Class for building dynamic structures, e.g. configuration structure which
    is customized for dynamic run time conditions.
    """

    def __init__(self, conditionals=None):
        """
        Initialise builder object. Set conditions and values.
        Parameters:
            conditionals   the conditional names and values to be used for
                           building
        Returns:
            None
        """
        self.conditionals = conditionals

    def build(self, structure):
        """
        Builds the result structure.
        Parameters:
            structure
        Returns:
            result structure
        """
        if isinstance(structure, dict):
            structure = structure.copy()
            for key in list(structure.keys()):
                if isinstance(key, Condition):
                    if key.evaluate(structure[key], self.conditionals):
                        # condition is met, delete the condition from dict
                        del structure[key]
                    else:
                        # condition is not met, drop the whole dict
                        return None

            for key in list(structure.keys()):
                if isinstance(key, Merge):
                    new = key.evaluate(structure[key], self.conditionals)
                    del structure[key]
                    new = self.build(new)
                    if new is None:
                        return None
                    else:
                        # for None, no merge needed, else merge:
                        structure.update(new)
                else:
                    # all other dictionary keys except Merge: bulid the
                    # sub-structure
                    new = self.build(structure[key])
                    if isinstance(structure[key], dict) and new is None:
                        # special case: if a dictionary becomes None, then a
                        # Condition failed inside --> delete the key from parent
                        del structure[key]
                    else:
                        structure[key] = new
        elif isinstance(structure, list):
            structure = list(structure)
            i = 0
            while i < len(structure):
                new = self.build(structure[i])
                if isinstance(structure[i], dict) and new is None:
                    # special case: if a dictionary becomes None, then a
                    # Condition failed inside --> delete the element from parent
                    del structure[i]
                else:
                    structure[i] = new
                    i += 1
        elif isinstance(structure, Substitute):
            new = structure.evaluate(self.conditionals)
            return self.build(new)
        return structure



"""
nilu/subprocess_helper.py
$Id: subprocess_helper.py 2824 2024-04-19 10:46:32Z pe $

Convenience functions for subprocess

History:
V.1.0.0  2012-12-11  pe  initial version

"""

import os
import shlex
import subprocess
import threading
import logging
import time
import six

class LogPipe(threading.Thread):
    """"
    Logger pipe class for connecting pythons subprocess module to the logging
    module. This is a helper for subprocess_log_streams().
    Povides a pipe that is constantly read in a new thread and written to a
    target logger.
    """
    def __init__(self, logger=None, level=None, stream=None, result=None):
        """
        Setup the LogPipe object, connect it to a given logger and a loglevel
        and start the thread to wait for pipe input.
        Parameters:
            logger    target logger to write the pipe's data to
            level     loglevel to be used
            result    additionally pass back the pipe's data as list of strings
        """
        threading.Thread.__init__(self)
        self.daemon = False
        self.logger = logger
        self.level = level
        self.stream = stream
        self.result = result
        self.fd_read, self.fd_write = os.pipe()
        self.pipe_reader = os.fdopen(self.fd_read)
        self.start()

    def fileno(self):
        """
        Return the write file descriptor of the pipe.
        """
        return self.fd_write

    def run(self):
        """
        Run the thread, logging everything that comes through the pipe.
        """
        for line in iter(self.pipe_reader.readline, ''):
            if self.result is not None:
                self.result.append(line)
            if self.logger is not None:
                self.logger.log(self.level, line.strip('\n'))
            if self.stream is not None:
                self.stream.write(line)
        self.pipe_reader.close()

    def close(self):
        """
        Close the write end of the pipe.
        """
        os.close(self.fd_write)

def subprocess_log_streams(args, logger=None,
                           stdoutlogger=None, stdoutlevel=logging.INFO,
                           stdoutstream=None,
                           stdoutresult=None,
                           stderrlogger=None, stderrlevel=logging.ERROR,
                           stderrstream=None,
                           stderrresult=None,
                           warntime=None, killtime=None):
    # pylint: disable=R0912,R0913
    #  R0912: Too many branches (13/12)
    #  R0913: Too many arguments
    """
    Forks a subprocess and logs it's stdout and stderr to (separate) loggers
    and/or writes to output streams.
    The function waits for the process to finish (needed for cleanup (closing
    write end of pipes).
    Parameters:
        args    arguments for subprocess (commandline as string or list of args
        logger  logger for logging messages of the caller side
        stdoutlogger  logger for subprocesses stdout
        stdoutlevel   loglevel for the subprocesses stdout
        stdoutstream  write stdout to output stream (file obj...)  
        stdoutresult  pass back stdout to caller (list.append)
        stderrlogger  logger for subprocesses stderr
        stderrlevel   loglevel for the subprocesses stderr
        stderrstream  write stderr to output stream (file obj...)  
        stderrresult  pass back stderr to caller (list.append)
        warntime      log warning to logger when runtime exceeds warntime (sec)
        killtime      kill subprocess when runtime exceeds killtime (seconds)
    This implementation is heavily inspired by:
    http://codereview.stackexchange.com/a/17959/20201 (thanks to deuberger)
    """
    if isinstance(args, six.string_types):
        args = shlex.split(args)
    if not logger:
        logger = logging.getLogger('subprocess')
    logger.info("spawning subprocess: " + ' '.join(args))

    stdoutpipe = None
    stderrpipe = None
    if stdoutlogger is not None or stdoutstream is not None or \
       stdoutresult is not None:
        stdoutpipe = LogPipe(logger=stdoutlogger, level=stdoutlevel,
                             stream=stdoutstream, result=stdoutresult)
    if stderrlogger is not None or stderrstream is not None or \
         stderrresult is not None:
        if stderrlogger == stdoutlogger and stderrlevel == stdoutlevel and \
           stderrstream == stdoutstream and stderrresult == stdoutresult:
            stderrpipe = subprocess.STDOUT
        else:
            stderrpipe = LogPipe(logger=stderrlogger, level=stderrlevel,
                                 stream=stderrstream, result=stderrresult)
    try:
        stin_pin, stin_pout = os.pipe()
        proc = subprocess.Popen(args, stdout=stdoutpipe, stderr=stderrpipe, stdin=stin_pin)
        logger.info("spawned pid: {0}".format(proc.pid))
        sec = 0
        while proc.poll() is None:
            if killtime and sec >= killtime:
                logger.error("subprocess still running - killing")
                proc.kill()
            if warntime and sec > 0 and sec % warntime == 0:
                logger.warn("subprocess still running - " + \
                            "parent is alive and waiting")
            time.sleep(1)
            sec += 1
        os.close(stin_pin)
        os.close(stin_pout)
    except OSError as expt:
        logger.error("error spawning process: {0}".format(expt.strerror))
        raise
    finally:
        if isinstance(stdoutpipe, LogPipe):
            stdoutpipe.close()
        if isinstance(stderrpipe, LogPipe):
            stderrpipe.close()
    logger.info("process {0} exited with code {1}"
                .format(proc.pid, proc.returncode))
    return proc



"""
$Id: vt.py 2088 2018-06-29 13:46:38Z pe $

EBAS Masterdata Class for volume std temperature masterdata

This module implements the class EbasMasterVT.

"""

from .offline_masterdata import VTOfflineMasterData
from .base import EbasMasterBase

class EbasMasterVT(EbasMasterBase, VTOfflineMasterData):
    """
    Domain Class for volume std temperature masterdata.
    Objects of this class do not represent entities, this class provides class
    methods for handling stations and checking them against master data.
    Stations master data are retrieved from database or from offline storage
    when no database access is possible.
    """

    # read offline masterdata for VT (volume std temperature)
    # Those are fallback values, will be read from database as soon as possible.
    VTOfflineMasterData.read_pickle_file()

    def __init__(self, dbh=None):
        """
        Read vt masterdata if dbh is provided.
        """
        VTOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)



"""
five: extensions to six module
pyhon 2+3=five
"""

import six

def raise_from(exc_value, exc_value_from):
    """
    Raise an exception from a context.
    Python 3 is handled well by six, python2 needs improvement.
    Parameters:
        exc_value       exception value for the exception to rise
        exc_value_from  causing exception, print context for this one
    Returns:
        None
    """

    if six.PY2:
        import traceback
        import sys
        traceback.print_exception(*sys.exc_info())
        sys.stderr.write(
            "\nThe above exception was the direct cause of the following "
            "exception:\n\n")
        raise exc_value
    six.raise_from(exc_value, exc_value_from)




"""
Basic functionality for handling parameters
(parameter groups, titles for sets of parameters)

Those methods are used both in ebas.io and ebas.domain.entities.ds.
In order to make them available for both modules they are separated in this
general domain module.
"""

from .dataset_types import is_auxiliary
from ..masterdata.pg_pl import EbasMasterPL, EbasMasterPG

EBAS_MASTER_PG = EbasMasterPG()
EBAS_MASTER_PL = EbasMasterPL()

def _optimise_groups(occ):
    """
    Optimise the use of component names and/or group names in order to describe
    the set of variables in the shortes possible way.
    Parameters:
        occ    occurance of different groups for each variable
               list of lists of tuples (list of variables, for each variable
               the list of possible groups and each of those is a tuple of
               (string [name], int [count of CO in the group]).
               The count of CO is set to 0 for comp_name (prefer comp name in
               case where only one variables is in a group)
    Result:
        list of groups/ components necessary to describe all variables (shortest
        possible combination of comp and groups)
    """
    # dictionary over all descriptions (keys), each element a list:
    #     [usable for # of variables, order number of occurances]
    descdict = {}
    for i in range(len(occ)):
        for j in range(len(occ[i])):
            if occ[i][j][0] not in descdict:
                descdict[occ[i][j][0]] = [1, occ[i][j][1]]
            else:
                descdict[occ[i][j][0]][0] += 1

    # convert to a list of tuples:
    #     (description, usable for # of variables, order number of
    #      occurances)
    # and sort the list: usable for # of variables (reversed!), order number
    # Thus, we get a list of descriptions in the order of preferred use:
    desclist = sorted(
        [(key, descdict[key][0], descdict[key][1])
         for key in descdict.keys()],
        key=lambda x: (-x[1], x[2]))

    # generate the result list: preferred description for each variable:
    res = []
    for i in range(len(occ)):
        vardescs = [x[0] for x in occ[i]]
        for desc in desclist:
            if desc[0] in vardescs:
                res.append(desc[0])
                break
    if len(res) != len(occ):
        # be safe
        raise RuntimeError()

    # return a set of all used descriptions, convert to sorted list
    return sorted(set(res))

def param_group(param_list):
    """
    Set of terms that describe the list of parameters in the shortest possible
    way (use group names and comp names)
    Parameters:
        param_list    list with parameter metadata (each tuple
                      (regime, matrix, comp_name))
    Returns:
        groups and/or components list for describing the parameters
    """
    all_groups = []
    for param in param_list:
        regime, matrix, comp_name = param
        if not is_auxiliary(matrix, comp_name):
            all_groups.append(
                [(comp_name, 0)] +
                [(x, EBAS_MASTER_PG[x].CO_COUNT)
                 for x in EBAS_MASTER_PL.list_groups(regime, matrix, comp_name)
                ])
            # Component name has score 0 (always use comp name if a group does
            # not lead to less expressions, i.e. only use a group if more
            # then one ds actually are in the group. See also below.
    # quickfix: if only aux...
    if not all_groups:
        for param in param_list:
            all_groups.append(
                [(comp_name, 0)] +
                [(x, EBAS_MASTER_PG[x].CO_COUNT)
                 for x in EBAS_MASTER_PL.list_groups(regime, matrix, comp_name)
                ])
    return _optimise_groups(all_groups)


def param_desc(param_list):
    """
    Set of terms that describe the list of parameters in the shortest possible
    way (use group descriptions (if not exists, group names) and comp names).
    Parameters:
        param_list    list with parameter metadata (each tuple
                      (regime, matrix, comp_name))
    Returns:
        description for parameters
    """
    return [group if group not in EBAS_MASTER_PG.META or
                not EBAS_MASTER_PG[group].PG_DESC
                else EBAS_MASTER_PG[group].PG_DESC
            for group in param_group(param_list)]

def title(param_list, station_list):
    """
    Generate a title for the series.
    Parameters:
        None
    Returns:
        description for parameters
    """
    desc = param_desc(param_list)
    stations = sorted(set(station_list))
    res = "{} at {}".format(
        ' and '.join([
            x for x in [', '.join(desc[:-1]), desc[-1]]
            if x]),
        ' and '.join([
            x for x in [', '.join(stations[:-1]), stations[-1]]
            if x]))
    if res.startswith('pH '):
        return res
    # first character uppercase
    return res[0:1].upper() + res[1:]



"""
ebas/domain/masterdata/ps.py
$Id: ps.py 2379 2019-11-19 09:58:18Z pe $

EBAS Masterdata Class for Persons

This module implements the class EbasMasterPS.
"""

from .. import EbasDomainError
from .offline_masterdata import PSOfflineMasterData
from .base import EbasMasterBase

class EbasMasterPS(EbasMasterBase, PSOfflineMasterData):
    """
    Domain Class for person masterdata.
    Objects of this class do not represent entities, this class provides class
    methods for handling persons and checking them against master data.
    Person master data are retrieved from database or from offline storage
    when no database access is possible.
    """

    # read offline masterdata for PS (persons)
    # Those are fallback values, will be read from database as soon as possible.
    PSOfflineMasterData.read_pickle_file()

    def __init__(self, dbh=None):
        """
        Read ps masterdata if dbh is provided.
        """
        PSOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)




"""
EBAS submission templates and template checks: NOx Template
"""

from .base import EbasTemplateBase, NoTemplate


class EbasTemplateNOxLev0(EbasTemplateBase):
    """
    Ebas template for NOx
    """

    TEMPLATE_NAME = 'NOx lev0'

    def match(self):
        """
        See if this template matches a file.
        """
        if (not self.file.metadata.datalevel or
                not self.file.metadata.datalevel.startswith('0')):
            raise NoTemplate()
        # Contains only NOx variables
        comp_names = [x.metadata.comp_name for x in self.file.variables]
        # identifying components (any of those makes the file a NOx file)
        ident_comps = ('NO_#counts', 'NO_converter_#counts', 'NO_sensitivity',
                       'nitrogen_monoxide', 'nitrogen_dioxide')
        # additional components
        add_comps = ('pressure', 'temperature', 'status',
                     'converter_efficiency')
        if not any([x in ident_comps for x in comp_names]):
            raise NoTemplate()
        if not all([x in ident_comps + add_comps for x in comp_names]):
            self.error(
                'no match because of additional variobles: {}'.format(
                    ', '.join([x for x in comp_names
                               if x not in ident_comps + add_comps])))
            raise NoTemplate()

        instr_types = set([self.file.get_meta_for_var(vnum, 'instr_type')
                      for vnum in range(len(self.file.variables))])
        if len(instr_types) != 1:
            self.error('only one instrumennt type for all variables allowed')
        self.instr_type = next(iter(instr_types))

        projs = self.file.metadata.projects
        strict_projs = ['ACTRIS', 'ACTRIS_NRT']
        self.actris = any([x in strict_projs for x in projs])
        self.actriserror = self.error if self.actris else self.warning

    def special_checks(self):
        """
        Special checks which cannot be done in the defailt checks.
        """
        if self.instr_type == 'chemiluminescence_molybdenum' and self.actris:
            self.error(
                'Instrument type {} not apporved by ACTRIS'.format(
                    self.instr_type))

        for vnum, var in enumerate(self.file.variables):
            comp = var.metadata.comp_name
            stat = self.file.get_meta_for_var(vnum, 'statistics')
            file_vnum, _, file_metalnum = self.file.msgref_vnum_lnum(vnum)
            if (comp in ('nitrogen_monoxide', 'nitrogen_dioxide') and
                    stat == 'arithmetic mean'):
                cal_scale = self.file.get_meta_for_var(vnum, 'cal_scale')
                if not cal_scale:
                    self.actriserror(
                        ("Variable {}: 'Calibration scale' metadata "
                         "missing").format(file_vnum), lnum=file_metalnum)
                stdtemp = self.file.get_meta_for_var(vnum, 'vol_std_temp')
                if not stdtemp:
                    self.actriserror(
                        ("Variable {}: 'Volume std. temperature' metadata "
                         "missing").format(file_vnum), lnum=file_metalnum)
                elif not isinstance(stdtemp, float):
                    self.actriserror(
                        ("Variable {}: 'Volume std. temperature' metadata "
                         "must be numeric").format(file_vnum),
                        lnum=file_metalnum)
                stdpres = self.file.get_meta_for_var(vnum, 'vol_std_pressure')
                if not stdpres:
                    self.actriserror(
                        ("Variable {}: 'Volume std. pressure' metadata "
                         "missing").format(file_vnum), lnum=file_metalnum)
                elif not isinstance(stdpres, float):
                    self.actriserror(
                        ("Variable {}: 'Volume std. pressure' metadata "
                         "must be numeric").format(file_vnum),
                        lnum=file_metalnum)
        checkvariables = [
            {
                'comp_name': 'pressure',
                'matrix': 'instrument',
                'characteristics': {
                    'Location': 'inlet',
                }
            },
            {
                'comp_name': 'pressure',
                'matrix': 'instrument',
                'characteristics': {
                    'Location': 'detector',
                }
            },
            {
                'comp_name': 'temperature',
                'matrix': 'instrument',
                'characteristics': {
                    'Location': 'inlet',
                }
            },
            {
                'comp_name': 'temperature',
                'matrix': 'instrument',
                'characteristics': {
                    'Location': 'detector',
                }
            },
            {
                'comp_name': 'status',
                'matrix': 'instrument',
                'characteristics': {
                    'Status type': 'calibration standard',
                }
            },
            {
                'comp_name': 'status',
                'matrix': 'instrument',
                'characteristics': {
                    'Status type': 'zero mode',
                }
            },
            {
                'comp_name': 'NO_#counts',
            },
            {
                'comp_name': 'NO_converter_#counts',
            },
            {
                'comp_name': 'NO_sensitivity',
            },
            {
                'comp_name': 'nitrogen_monoxide',
                'statistics': 'arithmetic mean'
            },
            {
                'comp_name': 'nitrogen_monoxide',
                'statistics': 'expanded uncertainty 2sigma'
            },
            {
                'comp_name': 'nitrogen_monoxide',
                'statistics': 'precision'
            },
            {
                'comp_name': 'nitrogen_monoxide',
                'statistics': 'detection limit'
            },
        ]
        checkvariables_converter = [
            {
                'comp_name': 'converter_efficiency',
            },
            {
                'comp_name': 'nitrogen_dioxide',
                'statistics': 'arithmetic mean'
            },
            {
                'comp_name': 'nitrogen_dioxide',
                'statistics': 'expanded uncertainty 2sigma'
            },
            {
                'comp_name': 'nitrogen_dioxide',
                'statistics': 'precision'
            },
            {
                'comp_name': 'nitrogen_dioxide',
                'statistics': 'detection limit'
            },
        ]
        if self.instr_type != 'chemiluminescence_photometer':
            checkvariables += checkvariables_converter

        # check if all variables are included (and appear exactly once)
        for elem in checkvariables:
            try:
                var = self.file.find_variable(elem)
            except ValueError as excpt:
                # variable is either missing, or more then one variable matches
                # "flatten" the characteristics:
                new = {x: elem[x] for x in elem if x != 'characteristics'}
                if 'characteristics' in elem:
                    new.update({x: elem['characteristics'][x]
                                for x in elem['characteristics']})
                self.actriserror("{}: {}".format(
                    str(excpt),
                    ', '.join(['='.join(
                        (self.file.internal.ebasmetadata.metadata_keys[x]['tag']
                         if x in self.file.internal.ebasmetadata.metadata_keys
                         else x, new[x])) 
                         for x in new.keys()])))
        # check for extrs variables:
        for vnum, unused_ in enumerate(self.file.variables):
            file_vnum, _, file_metalnum = self.file.msgref_vnum_lnum(vnum)
            if not any([self.file.match_variable(vnum, elem)
                        for elem in checkvariables]):
                # variable is not allowed
                self.error(
                    "Variable {}:  variable is not allowed".format(file_vnum),
                    lnum=file_metalnum)

        # check mandatory metadata:
        mandatory_global_metadata = {
            'ozone_corr': 'Not corrected for reaction with O3 in the inlet',
            'watervapor_corr': 'Not corrected for water vapor quenching in CLD',
        }
        if self.instr_type != 'chemiluminescence_photometer':
            mandatory_global_metadata.update({
                'time_inlet_to_converter': None,
                'time_converter_or_bypass_line': None,
                'time_stay_converter': None,
                'converter_temp': None,
            })
        for meta in mandatory_global_metadata:
            if not self.file.metadata[meta]:
                self.actriserror(
                    "Mandatory global metadata element '{}' missing".format(
                        self.file.internal.ebasmetadata.metadata_keys[meta]['tag']))
            elif (mandatory_global_metadata[meta] is not None and
                    self.file.metadata[meta] !=
                        mandatory_global_metadata[meta]):
                self.actriserror(
                    "Global metadata element '{}' must be '{}'".format(
                        self.file.internal.ebasmetadata.metadata_keys[meta]['tag'],
                        mandatory_global_metadata[meta]))

class EbasTemplateNOxLev1(EbasTemplateBase):
    """
    Ebas template for NOx
    """

    TEMPLATE_NAME = 'NOx lev1'

    def match(self):
        """
        See if this template matches a file.
        """
        if (not self.file.metadata.datalevel or
                not self.file.metadata.datalevel.startswith('1')):
            raise NoTemplate()
        # Contains only NOx variables
        comp_names = [x.metadata.comp_name for x in self.file.variables]
        # identifying components (any of those makes the file a NOx file)
        ident_comps = ('nitrogen_monoxide', 'nitrogen_dioxide', 'NOx')
        # additional components
        add_comps = ('pressure', 'temperature')
        if not any([x in ident_comps for x in comp_names]):
            raise NoTemplate()
        if not all([x in ident_comps + add_comps for x in comp_names]):
            self.error(
                'no match because of additional variobles: {}'.format(
                    ', '.join([x for x in comp_names
                               if x not in ident_comps + add_comps])))
            raise NoTemplate()

        instr_types = set([self.file.get_meta_for_var(vnum, 'instr_type')
                      for vnum in range(len(self.file.variables))])
        if len(instr_types) != 1:
            self.error('only one instrumennt type for all variables allowed')
        self.instr_type = next(iter(instr_types))

        projs = self.file.metadata.projects
        strict_projs = ['ACTRIS', 'ACTRIS_NRT']
        self.actris = any([x in strict_projs for x in projs])
        self.actriserror = self.error if self.actris else self.warning

    def special_checks(self):
        """
        Special checks which cannot be done in the defailt checks.
        """
        if self.instr_type == 'chemiluminescence_molybdenum' and self.actris:
            self.error(
                'Instrument type {} not apporved by ACTRIS'.format(
                    self.instr_type))

        for vnum, var in enumerate(self.file.variables):
            comp = var.metadata.comp_name
            stat = self.file.get_meta_for_var(vnum, 'statistics')
            file_vnum, _, file_metalnum = self.file.msgref_vnum_lnum(vnum)
            if (comp in ('nitrogen_monoxide', 'nitrogen_dioxide', 'NOx') and
                    stat == 'arithmetic mean'):
                cal_scale = self.file.get_meta_for_var(vnum, 'cal_scale')
                if not cal_scale and comp != 'NOx':
                    # TODO: shouldn't NOX have Cal. scale NPL+GPT?
                    self.actriserror(
                        ("Variable {}: 'Calibration scale' metadata "
                         "missing").format(file_vnum), lnum=file_metalnum)
                stdtemp = self.file.get_meta_for_var(vnum, 'vol_std_temp')
                if not stdtemp:
                    self.actriserror(
                        ("Variable {}: 'Volume std. temperature' metadata "
                         "missing").format(file_vnum), lnum=file_metalnum)
                elif not isinstance(stdtemp, float):
                    self.actriserror(
                        ("Variable {}: 'Volume std. temperature' metadata "
                         "must be numeric").format(file_vnum),
                        lnum=file_metalnum)
                stdpres = self.file.get_meta_for_var(vnum, 'vol_std_pressure')
                if not stdpres:
                    self.actriserror(
                        ("Variable {}: 'Volume std. pressure' metadata "
                         "missing").format(file_vnum), lnum=file_metalnum)
                elif not isinstance(stdpres, float):
                    self.actriserror(
                        ("Variable {}: 'Volume std. pressure' metadata "
                         "must be numeric").format(file_vnum),
                        lnum=file_metalnum)
        checkvariables = [
            {
                'comp_name': 'pressure',
                'matrix': 'instrument',
                'characteristics': {
                    'Location': 'inlet',
                }
            },
            {
                'comp_name': 'temperature',
                'matrix': 'instrument',
                'characteristics': {
                    'Location': 'inlet',
                }
            },
            {
                'comp_name': 'nitrogen_monoxide',
                'statistics': 'arithmetic mean'
            },
            {
                'comp_name': 'nitrogen_monoxide',
                'statistics': 'expanded uncertainty 2sigma'
            },
            {
                'comp_name': 'nitrogen_monoxide',
                'statistics': 'precision'
            },
            {
                'comp_name': 'nitrogen_monoxide',
                'statistics': 'detection limit'
            },
        ]
        checkvariables_converter = [
            {
                'comp_name': 'nitrogen_dioxide',
                'statistics': 'arithmetic mean'
            },
            {
                'comp_name': 'nitrogen_dioxide',
                'statistics': 'expanded uncertainty 2sigma'
            },
            {
                'comp_name': 'nitrogen_dioxide',
                'statistics': 'precision'
            },
            {
                'comp_name': 'nitrogen_dioxide',
                'statistics': 'detection limit'
            },
            {
                'comp_name': 'NOx',
                'statistics': 'arithmetic mean'
            },
            {
                'comp_name': 'NOx',
                'statistics': 'expanded uncertainty 2sigma'
            },
            {
                'comp_name': 'NOx',
                'statistics': 'precision'
            },
            {
                'comp_name': 'NOx',
                'statistics': 'detection limit'
            },
        ]
        if self.instr_type != 'chemiluminescence_photometer':
            checkvariables += checkvariables_converter

        # check if all variables are included (and appear exactly once)
        for elem in checkvariables:
            try:
                var = self.file.find_variable(elem)
            except ValueError as excpt:
                # variable is either missing, or more then one variable matches
                # "flatten" the characteristics:
                new = {x: elem[x] for x in elem if x != 'characteristics'}
                if 'characteristics' in elem:
                    new.update({x: elem['characteristics'][x]
                                for x in elem['characteristics']})
                self.actriserror("{}: {}".format(
                    str(excpt),
                    ', '.join(['='.join(
                        (self.file.internal.ebasmetadata.metadata_keys[x]['tag']
                         if x in self.file.internal.ebasmetadata.metadata_keys
                         else x, new[x])) 
                         for x in new.keys()])))
        # check for extrs variables:
        for vnum, unused_ in enumerate(self.file.variables):
            file_vnum, _, file_metalnum = self.file.msgref_vnum_lnum(vnum)
            if not any([self.file.match_variable(vnum, elem)
                        for elem in checkvariables]):
                # variable is not allowed
                self.error(
                    "Variable {}:  variable is not allowed".format(file_vnum),
                    lnum=file_metalnum)

        # check mandatory metadata:
        mandatory_global_metadata = {
            'ozone_corr': 'Not corrected for reaction with O3 in the inlet',
            'watervapor_corr': 'Not corrected for water vapor quenching in CLD',
        }
        #if self.instr_type != 'chemiluminescence_photometer':
        #    mandatory_global_metadata.update({
        #    })
        for meta in mandatory_global_metadata:
            if not self.file.metadata[meta]:
                self.actriserror(
                    "Mandatory global metadata element '{}' missing".format(
                        self.file.internal.ebasmetadata.metadata_keys[meta]['tag']))
            elif (mandatory_global_metadata[meta] is not None and
                    self.file.metadata[meta] !=
                        mandatory_global_metadata[meta]):
                self.actriserror(
                    "Global metadata element '{}' must be '{}'".format(
                        self.file.internal.ebasmetadata.metadata_keys[meta]['tag'],
                        mandatory_global_metadata[meta]))



"""
/Xml/base.py
$Id: ebas_diffexport.py 2466 2020-07-10 13:20:48Z pe $

Base class for XML.
Has (still) all attributes as base class for Nasa Ames 1001.

History:
V.1.0.0  2015-02-16  pe   initial version

"""
import logging
from .generateDS_ebas_diffexport import ebas_diffexport
import sys
import os
import pytz
from six import string_types

class DiffXmlError(Exception):
    """
    Error class for XML.
    """
    pass

class DiffXmlFileExists(DiffXmlError):
    """
    Exception class raised when the DB file already exists.
    """
    pass

class EbasDiffExportXML(object):
    """
    EbasDiffExportXML
    """

    def __init__(self, db_state, diff_state, xmlfile):
        """
        Initialize diffExport xml object
        Paramaters:
            db_state     database state for export. datetime.datetime
            diff_state   diff state for database export datetime.datetime
            Both parameters must be either tz naive or tz=UTC (if naive, time
            zone is set to UTC)
        Returns:
            None
        Raises:
            IOError            if xmlfile cannot be opened
            DiffXmlFileExists  if xmlfile exists already
            DiffXmlError       on other errors
        """
        self.logger = logging.getLogger("diffExportXML")
        if xmlfile is None or xmlfile == '-':
            self.file = sys.stdout
            self.logger.debug("writing diffxml to stdout")
        elif isinstance(xmlfile, string_types):
            if os.path.exists(xmlfile):
                raise DiffXmlFileExists('file {} exists'.format(xmlfile))
            self.logger.debug("writing diffxml to file %s", xmlfile)
            self.file = open(xmlfile, "w")
        else:
            self.file = xmlfile
            self.logger.debug("writing diffxml to stream")

        if db_state.tzinfo and db_state.tzinfo != pytz.timezone('UTC'):
            raise DiffXmlError('time zone: only UTC or naive allowed')
        elif not db_state.tzinfo:
            db_state = pytz.utc.localize(db_state)
        if diff_state.tzinfo and diff_state.tzinfo != pytz.timezone('UTC'):
            raise DiffXmlError('time zone: only UTC or naive allowed')
        elif not diff_state.tzinfo:
            diff_state = pytz.utc.localize(diff_state)

        self.diff_export = ebas_diffexport.diffExportType(dbState=db_state,
                                                          diffState=diff_state)



    def add_deleted(self, dataset_key, start_time, end_time):
        """
        Adds a deleted interval entry.
        Parameters:
            datasetKey     the ds_setkey for the data
            startTime      start time of the deleted interval
            endTime        end time of the deleted interval
        Returns:
            None
        """
        if start_time.tzinfo and start_time.tzinfo != pytz.timezone('UTC'):
            raise DiffXmlError('time zone: only UTC or naive allowed')
        elif not start_time.tzinfo:
            start_time = pytz.utc.localize(start_time)
        if end_time.tzinfo and end_time.tzinfo != pytz.timezone('UTC'):
            raise DiffXmlError('time zone: only UTC or naive allowed')
        elif not end_time.tzinfo:
            end_time = pytz.utc.localize(end_time)
        deleted = ebas_diffexport.dataIntervalType(datasetKey=dataset_key,
                                                   startTime=start_time,
                                                   endTime=end_time)
        self.diff_export.add_deleted(deleted)

    def add_added(self, dataset_key, start_time, end_time):
        """
        Adds an added interval entry.
        Parameters:
            datasetKey     the ds_setkey for the data
            startTime      start time of the added interval
            endTime        end time of the added interval
        Returns:
            None
        """
        if start_time.tzinfo and start_time.tzinfo != pytz.timezone('UTC'):
            raise DiffXmlError('time zone: only UTC or naive allowed')
        elif not start_time.tzinfo:
            start_time = pytz.utc.localize(start_time)
        if end_time.tzinfo and end_time.tzinfo != pytz.timezone('UTC'):
            raise DiffXmlError('time zone: only UTC or naive allowed')
        elif not end_time.tzinfo:
            end_time = pytz.utc.localize(end_time)
        added = ebas_diffexport.dataIntervalType(datasetKey=dataset_key,
                                                 startTime=start_time,
                                                 endTime=end_time)
        self.diff_export.add_added(added)

    def write(self):
        """
        Writes the file.
        Parameters:
            filespec    file name (incl. path) or file like object (stream)
                        stdout if not passed
        Returns:
            None
        """
        self.diff_export.export(self.file, 0, name_='diffExport')
        self.file.close()




"""
Basic file IO for ASCII datafiles. Mainly encoding and special character issues.
"""

import logging
import os
from six import string_types, PY2
from nilutility.file_helper import tail


class AsciiReadError(Exception):
    """
    Error class for NasaAmes 1001.
    """
    pass

class BackposError(Exception):
    """
    Exception class when backwards positioning in file was not allowed.
    """
    pass

class AsciiRead(object):
    # pylint: disable=R0902
    # R0902: Too many instance attributes
    """
    Basic file IO for ASCII datafiles. Mainly encoding and special character
    issues.
    The instances of this class are a context managers, so they can be used in
    a "with" block.
    """

    def __init__(
            self, filespec,
            encodings=('ascii', 'utf-8', 'utf-16', 'utf-32'),
            fallback_encoding='iso8859-15', strict_enc=True,
            ignore_bom=1, cvt_nbsp=False, strip=False, tabsize=None,
            error_callback=None, warning_callback=None):
        # pylint: disable=R0913
        # R0913: Too many arguments
        """
        Set up IO object.
        Opens the underlying file right away, this way callers can easier catch
        an IOError if a "with" block schould be used.
        Example:
        try:
           fil = AsciiRead(filename)
        except IOEror:
           ...
        with fil:
           ...

        Parameters:
            filespec    either a path and filename or a stream like object
            encodings   list of non-ambiguous encodings to try
            fallback_encoding
                        a fallback 8 bit encoding, this is ambiguous and
                        might be correct or not
            strict_enc  be strict about encodings (do not allow changes through
                        the file)
            ignore_bom  ignore (take out) BOM character in utf- encoded files
                        possible values:
                        1   take out as first character in file
                        2   take out as any character in file
            cvt_nbsp    convert non breaking space characters to normal space
                        1 convert, but raise warning only first time
                        2 convert, warn each time
                        3 convert, no warning
            strip       strip leading and trailing white spaces (and \n)
            tabsize     convert tab to newlines, use tabsize
                        None = no tab conversion
                        If tabsize is negative use fixed number of blanks for
                        each tab instead og expandtabs.
            error_callback
                        Callback method in case of error. See method error.
            warning_callback
                        Callback method in case of warning. See method warning.
            None
        Returns:
            None
        Raises:
            None
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.filespec = filespec
        self.filename = None
        self.file = None
        self.lnum = 0
        self.encodings = encodings
        self.fallback_encoding = fallback_encoding
        self.encoding = None
        self.enc_line = None
        self.strict_enc = strict_enc
        self.ignore_bom = ignore_bom
        self.cvt_nbsp = cvt_nbsp
        self.strip = strip
        self.tabsize = tabsize
        self.error_callback = error_callback
        self.warning_callback = warning_callback
        # Open the file right away. See docstring for explanation
        self.open()

    def __enter__(self):
        """
        Context manager enter method.
        Parameters:
            None
        Returns:
            self as context manager object
        """
        return self

    def __exit__(self,
                 exception_type, exception_value, traceback): # @UnusedVariable
        """
        Context manager exit method.
        Parameters:
            exception_type, exception_value, traceback
                exception details (if an exception is raised in the with block)
        Returns:
            None (= False, True would suppress the exception)
        """
        self.close()

    def open(self):
        """
        Opens the file.
        Parameters:
        Returns:
            None
        """
        if PY2:
            try:
                import pathlib
            except ImportError:
                accepted = string_types
            else:
                accepted = (string_types, pathlib.PurePath)
        else:
            accepted = (string_types, os.PathLike)
        if isinstance(self.filespec, accepted):
            self.filename = self.filespec
            if PY2:
                self.file = open(self.filename, "Urb")
            else:
                self.file = open(self.filename, "r")
            self.logger.info("reading file %s", self.filename)
        else:
            self.filename = None
            self.file = self.filespec
            self.logger.info("reading from stream")

    def close(self):
        """
        Closes the file.
        Parameters:
            None
        Returns:
            None
        """
        if self.filename:
            # only close file if we opened it in open()
            self.file.close()

    def readline(self):
        """
        Read one line of data. Get the net content (no line breaks).

        Parameters:
        """
        self.lnum += 1
        rawline = self.file.readline()
        if rawline == "":
            self.logger.debug(u"read line %d: EOF", self.lnum)
            raise EOFError()
        self.logger.debug(u"read raw line %d: %s", self.lnum, repr(rawline))
        if PY2:
            line = self.decode(rawline)
        else:
            line = rawline
        if self.cvt_nbsp and u'\u00a0' in line:
            if self.cvt_nbsp == 1:
                self.warning('File contains NO-BREAK SPACE characters, will be '
                             'substituted with normal blanks')
                line = line.replace(u'\u00a0', ' ')
                self.cvt_nbsp = 3
            elif self.cvt_nbsp == 2:
                while u'\u00a0' in line:
                    pos = line.index(u'\u00a0') + 1
                    self.warning('Pos {}: NO-BREAK SPACE character, will be '
                                 'substituted with normal blank'.format(pos))
                    line = line.replace(u'\u00a0', ' ', 1)
            else:
                line = line.replace(u'\u00a0', ' ')
        if self.lnum == 1 and self.ignore_bom and line.startswith(u'\ufeff'):
            if self.encoding == 'utf-8':
                self.warning(
                    "File starts with unicode Bit Order Marker (BOM) "
                    "character. "
                    "This is neither needed nor recommended in utf-8 encoding. "
                    "BOM will be ignored.")
            line = line.replace(u"\ufeff", "", 1)
        while self.ignore_bom == 2 and u'\ufeff' in line and \
                (self.lnum > 1 or not line.startswith(u'\ufeff')):
            pos = line.index(u'\ufeff') + 1
            self.warning(
                "Pos {}: Unicode Bit Order Marker (BOM) character at unusual "
                "position (not the start of the file). "
                "BOM will be ignored.".format(pos))
            line = line.replace(u"\ufeff", "", 1)
        if self.strip:
            # strip leading and trailing whitespaces, \n
            line = line.strip()
        if self.tabsize is not None and self.tabsize >= 0:
            # if tabsize is 0 or positive, use expandtabs
            line = line.expandtabs(self.tabsize)
        elif self.tabsize:
            # if tabsize is negative, use fixed number of blanks for each tab
            line = line.replace(u"\t", " " * (-1 * self.tabsize))
        self.logger.debug(u"encoded line %d: %s", self.lnum, repr(line))
        return line

    def decode(self, line):
        """
        Decode a single line of input.
        Parameters:
            line     line of input from the file
        Retruns:
            decoded line (unicode)
        """
        # Try first 7 bit ascii, then the preselected encodings (self.encoding)
        # if set.
        # If strict_enc is not set, select a different encoding if the
        # preselected does not work.

        # If no encoding is preselected (self.encoding is None), then the
        # selection of a suitable codecs will be started.
        # As long as only 7 bit ascii characters are used (codec 'ascii'), there
        # is no change (self.encoding is still None, which means 'whatever may
        # come later').
        # As soon as a UnicodeDecodeError is raised, the diffeent encodings are
        # tried in order.
        # When one of those is selected, self.encoding is fixed, from now on
        # this codec is used for reading the file. If self.strict_enc is set,
        # This encoding must be used for the rest of the file (no change).
        #
        # The different encodings will raise NO warning, because they are
        # officially supported and unambiguous character sets. Fallback to]
        # fallback_encoding will raise a warning (ambiguous encoding).
        #
        # There can be only ONE eight bit encoding (fallback_encoding), because
        # single byte encodings will always succeed (although maybe producing
        # nonsense)
        #
        # utf-8 and single byte encodings:
        # The large number of invalid byte sequences [in utf-8] provides the
        # advantage of making it easy to have a program accept both UTF-8 and
        # legacy encodings such as ISO-8859-1. Thus, the software can check for
        # UTF-8 correctness, and if that fails assume the input to be in the\
        # legacy encoding. It is technically true that this may detect an
        # ISO-8859-1 string as UTF-8, but this is very unlikely if it contains
        # any 8-bit bytes as they all have to be in unusual patterns of two or
        # more in a row, such as ...
        # http://en.wikipedia.org/wiki/UTF-8#Invalid_byte_sequences

        if self.encoding:
            try:
                return line.decode(self.encoding)
            except UnicodeDecodeError:
                if not self.strict_enc:
                    self.encoding = None
                    return self.decode(line)
                if self.enc_line != None:
                    self.error(
                        'character set decode error: line {} suggested codec '
                        "'{}', but it failed in line {}"
                        .format(self.enc_line, self.encoding, self.lnum))
                else:
                    self.error(
                        "character set decode error: codec '{}' failed in line "
                        '{}'.format(self.encoding, line))
        else:
            try:
                return line.decode('ascii')
            except UnicodeDecodeError:
                for enc in range(len(self.encodings)):
                    try:
                        line = line.decode(self.encodings[enc])
                        self.encoding = self.encodings[enc]
                        self.enc_line = self.lnum
                        return line
                    except UnicodeDecodeError:
                        pass
                if self.fallback_encoding:
                    try:
                        line = line.decode(self.fallback_encoding)
                    except UnicodeDecodeError:
                        self.error(
                            "unsupported character encoding; tried: {}".format(
                                ('ascii',) +
                                self.encodings +
                                (self.fallback_encoding,)
                                if self.fallback_encoding
                                else tuple()))
                    self.encoding = self.fallback_encoding
                    self.enc_line = self.lnum
                    self.warning(
                        "character set encoding: fallback to codec '{}'"
                        ' but this might be wrong'
                        .format(self.fallback_encoding))
                    return line

    def goto_line(self, lnum, allow_backpos=True):
        """
        Position the file pointer to specified line number.
        Parameters:
            lnum   line number (positive=from start, negative=from end)
                   0 = before reading the first line
                   -2 = before reading the last line
        Returns:
            None
        Raises:
            EOFError when reaching or exceeding end of file
            BackposError when allow_backpos was False and backwards positioning
            occurred
        """
        oldpos = self.file.tell()
        oldlnum = self.lnum
        if lnum >= 0:
            if self.lnum > lnum:
                # if we are already past, start from the start
                self.file.seek(0, os.SEEK_SET)
                self.lnum = 0
            while self.lnum < lnum:
                self.readline()
        else:
            lines = tail(self.file, (-1) * lnum - 1)
            # seek to length of last n lines
            self.file.seek(0, os.SEEK_END)
            self.file.seek(self.file.tell() - sum([len(l) for l in lines]),
                           os.SEEK_SET)
        if self.file.tell() < oldpos:
            self.file.seek(oldpos, os.SEEK_SET)
            self.lnum = oldlnum
            if not allow_backpos:
                raise BackposError()

    def error(self, msg):
        """
        Error handling for file read.

        Error callback method (self.error_callback):
        If the error callback method is set (self.error_callback), it is called.
        The error callback method must accept following parameters:
            msg
        If the error callback returns True, the error is logged and the
        exception raised if needed.
        If the error callback returns False, everything will be handled by the
        callback method and the local error method will cease action.

        Parameters:
            msg         error message
        Returns:
            None
        Raises:
            NasaAmesError when parameter exception is True
        """
        if not self.error_callback or \
                self.error_callback(msg):
            msg = "line {}: {}".format(self.lnum, msg)
            self.logger.error(msg)
            raise AsciiReadError(msg)

    def warning(self, msg):
        """
        Warning message for file read.

        Warning callback method (self.error_callback):
        If the warning callback method is set (self.warning_callback), it is
        called.
        The warning callback method must accept following parameters:
            msg
        If the warning callback returns True, the warning is logged and the
        exception raised if needed.
        If the warning callback returns False, everything will be handled by the
        callback method and the local warning method will cease action.

        Parameters:
            msg         warning message
        Returns:
            None
        """
        if not self.warning_callback or \
                self.warning_callback(msg):
            msg = "line {}: {}".format(self.lnum, msg)
            self.logger.warning(msg)



"""
ebas.domain.basic_domain_logic.flags
$Id: flags.py 2561 2020-12-07 23:09:30Z pe $
Basic ebas domain logic for flags.
"""

from ebas.domain.masterdata.fl import EbasMasterFL
from .base import EbasDomainIssueList, \
    ISSUE_SEVERITY_ERROR, ISSUE_SEVERITY_WARNING, \
    MSGID_FLAGCHECK_ILLEGAL, MSGID_FLAGCHECK_MISSING_100, \
    MSGID_FLAGCHECK_VALID_VALUE_MISSIG_FLAG, \
    MSGID_FLAGCHECK_MISSING_VALUE_FLAG_100, \
    MSGID_FLAGCHECK_MISSING_VALUE_NO_FLAG, \
    MSGID_FLAGCHECK_PRECIP_FLAG_782_783_784_DROPPED, \
    MSGID_FLAGCHECK_PRECIP_FLAG_782_783_784_ERROR, \
    MSGID_FLAGCHECK_PRECIP_FLAG_890_DROPPED, \
    MSGID_FLAGCHECK_PRECIP_FLAG_890_SET_ZERO, \
    MSGID_FLAGCHECK_PRECIP_FLAG_890_ERROR, \
    MSGID_FLAGCHECK_AUX_VALID_VALUE_MISSING_FLAG

def get_flag_summary(flags, want_issues=True):
    """
    Get a summary of flag validity.
    Parameters:
        flags        list of flags for one sample (list of int)
        want_issues  bool, True if a list of issues should be returned
                     (not needed for export from DB...)
    Returns:
        tupel (valid, invalid, missing, issues) if issues == True
        else tupel (valid, invalid, missing)
        numbers for invalid, valid and missing flags
        as well as list of issues (EbasDomainIssueList)
    """
    emfl = EbasMasterFL()
    valid, invalid, missing = 0, 0, 0
    if want_issues:
        issues = EbasDomainIssueList(
            keys=('severity', 'msg_id', 'message'),
            msg_id=(MSGID_FLAGCHECK_ILLEGAL, MSGID_FLAGCHECK_MISSING_100))
    for flag in flags:
        try:
            validity = emfl[flag]['FL_VALIDITY']
        except KeyError:
            if want_issues:
                issues.add(
                    ISSUE_SEVERITY_ERROR, MSGID_FLAGCHECK_ILLEGAL,
                    "illegal flag: {0}".format(flag))
            continue
        if validity == 'V' or validity == 'H':
            valid += 1
        if validity == 'I':
            invalid += 1
        elif validity == 'M':
            missing += 1
    if 100 in flags:
        if missing > 1 and want_issues:
            issues.add(
                ISSUE_SEVERITY_ERROR, MSGID_FLAGCHECK_MISSING_100,
                "invalid combination of missing flag and flag 100")
        # force re-validation of invalid flagged data
        # data must not be missing!
        invalid = 0
    return (valid, invalid, missing, issues) if want_issues \
        else (valid, invalid, missing)

def check_value_flag_consistency(
        values, flags, ds_special, ignore_flagconsistency, fix_flagconsistency):
    """
    Checks consistency of values and flags for samples.
    Does not directly log issues but collects issues in a EbasDomainIssueList.
    Thus, the caller can employ a message condenser object to condense messages
    to be logged.
    Parameters:
        values     list of values for a time series
        flags      list fo flags (list of lists) of a time series
        ds_special special type of data
                    1 precipitation amount
                    2 auxiliary data
                      (special case: delete value if missing flag)
    Returns:
        EbasDomainIssueList with issues found (empty list if OK)
    """
    issues = EbasDomainIssueList()
    if len(values) != len(flags):
        RuntimeError('Values and flag lists not equal size')
    for i in range(len(values)):  # pylint: disable=C0200
        # C0200: Consider using enumerate instead of iterating with range/len
        # --> only using index
        # change value, drop some flags in advance:
        if ds_special == 1:  # precipitation amount
            _check_value_flag_precip_amount(
                values, flags, ignore_flagconsistency, fix_flagconsistency, i,
                issues)
        if ds_special == 2:  # auxiliary data
            _check_value_flag_auxiliary_ds(
                values, flags, ignore_flagconsistency, fix_flagconsistency, i,
                issues)
        _, invalid, missing, flag_issues = get_flag_summary(flags[i])
        for err in flag_issues:
            issues.add(err['severity'], err['msg_id'], i,
                       err['message'] + " value={}, flags={}".format(
                           values[i], flags[i]))
        if missing > 0 and values[i] != None:
            sev = ISSUE_SEVERITY_ERROR
            msg = ("value/flag inconsistent: value={}, flags={}; valid "
                   "value and missing flag")
            if fix_flagconsistency:
                sev = ISSUE_SEVERITY_WARNING
                values[i] = None
                msg += "; Setting value to MISSING (fix_flagconsistency)"
            elif ignore_flagconsistency:
                sev = ISSUE_SEVERITY_WARNING
            issues.add(
                sev, MSGID_FLAGCHECK_VALID_VALUE_MISSIG_FLAG, i, msg.format(
                    values[i], flags[i]))
        if values[i] is None and 100 in flags[i]:
            sev = ISSUE_SEVERITY_ERROR
            msg = ("value/flag inconsistent: value={}, flags={}; missing "
                   "value and flag 100 is not allowed")
            if fix_flagconsistency:
                sev = ISSUE_SEVERITY_WARNING
                flags[i] = [x for x in flags[i] if x != 100]
                msg += "; Removing flag 100 (fix_flagconsistency)"
            elif ignore_flagconsistency:
                sev = ISSUE_SEVERITY_WARNING
            issues.add(
                sev, MSGID_FLAGCHECK_MISSING_VALUE_FLAG_100, i, msg.format(
                    values[i], flags[i]))
        elif missing == 0 and invalid == 0 and values[i] is None:
            issues.add(
                ISSUE_SEVERITY_WARNING, MSGID_FLAGCHECK_MISSING_VALUE_NO_FLAG,
                i,
                "value/flag inconsistent: value={}, flags={}; no "
                "missing or invalid flag: flag 999 added"
                .format(values[i], flags[i]))
            flags[i].append(999)
            missing += 1
    return issues

def _check_value_flag_precip_amount(
    values, flags, ignore_flagconsistency, fix_flagconsistency, idx, issues):
    """
    Special consistency checks and fixes for precipitation amount data.
    Parameters:
        values     list of values for a time series
        flags      list fo flags (list of lists) of a time series
        idx        index of sample to be checked/fixed
        issues     issues list
    Returns:
        None
    Raises:
        None
    """
    for flag in (782, 783, 784):
        # 784    I    Low precipitation, concentration estimated
        # 783    M    Low precipitation, concentration unknown
        # 782    V    Low precipitation, concentration estimated
        if flag in flags[idx]:
            if values[idx] != None:
                issues.add(
                    ISSUE_SEVERITY_WARNING,
                    MSGID_FLAGCHECK_PRECIP_FLAG_782_783_784_DROPPED, idx,
                    "invalid flag {} for precipitation_amount; dropped".format(
                        flag))
                while flag in flags[idx]:
                    flags[idx].remove(flag)
            else:
                sev = ISSUE_SEVERITY_ERROR
                msg = ("invalid flag {} for precipitation_amount with MISSING "
                       "value")
                if fix_flagconsistency:
                    sev = ISSUE_SEVERITY_WARNING
                    msg += "; Removing flag {} (fix_flagconsistency)".format(
                        flag)
                    while flag in flags[idx]:
                        flags[idx].remove(flag)
                elif ignore_flagconsistency:
                    sev = ISSUE_SEVERITY_WARNING
                issues.add(
                    sev, MSGID_FLAGCHECK_PRECIP_FLAG_782_783_784_ERROR, idx,
                    msg.format(flag))
    if 890 in flags[idx]:
        # 890    M    Concentration in precipitation undefined, no
        #             precipitation
        if set(flags[idx]) == {890} and values[idx] == 0.0:
            issues.add(
                ISSUE_SEVERITY_WARNING, MSGID_FLAGCHECK_PRECIP_FLAG_890_DROPPED,
                idx,
                "invalid flag 890 for precipitation amount; "
                "dropping flag 890")
            while 890 in flags[idx]:
                flags[idx].remove(890)
        elif set(flags[idx]) == {890} and values[idx] is None:
            issues.add(
                ISSUE_SEVERITY_WARNING,
                MSGID_FLAGCHECK_PRECIP_FLAG_890_SET_ZERO,
                idx,
                "invalid flag 890 for precipitation amount; "
                "changing MISSING value to 0.0, dropping flag 890")
            while 890 in flags[idx]:
                flags[idx].remove(890)
            values[idx] = 0.0
        else:
            # fix_flagconsistency? Can we fix this?
            sev = ISSUE_SEVERITY_ERROR
            if fix_flagconsistency or ignore_flagconsistency:
                sev = ISSUE_SEVERITY_WARNING
            issues.add(
                sev, MSGID_FLAGCHECK_PRECIP_FLAG_890_ERROR, idx,
                "invalid flag 890 for precipitation amount")

def _check_value_flag_auxiliary_ds(
    values, flags, ignore_flagconsistency, fix_flagconsistency, idx, issues):
    """
    Special consistency checks and fixes for auxiliary datasets.
    Parameters:
        values     list of values for a time series
        flags      list fo flags (list of lists) of a time series
        idx        index of sample to be checked/fixed
        issues     issues list
    Returns:
        None
    Raises:
        None
    """
    _, _, missing, _ = get_flag_summary(flags[idx])
    if missing > 0 and values[idx] != None:
        emfl = EbasMasterFL()
        rem = []
        for flg in flags[idx]:
            try:
                if emfl[flg]['FL_VALIDITY'] == 'M':
                    rem.append(flg)
            except KeyError:
                # non existing flag, ignore here, will be error in
                # general check_value_flag_consistency
                continue
        issues.add(
            ISSUE_SEVERITY_WARNING,
            MSGID_FLAGCHECK_AUX_VALID_VALUE_MISSING_FLAG,
            idx,
            "value/flag inconsistent: value={}, flags={}; valid value and "
            "missing flag{}; fix for auxiliary data: remove missing flag{} "
            "({})".format(
                values[idx], flags[idx], 's' if len(rem) > 1 else '',
                's' if len(rem) > 1 else '', ', '.join([str(x) for x in rem])))
        for fl_ in rem:
            flags[idx].remove(fl_)



"""
/Xml/write.py
$Id: write.py 2466 2020-07-10 13:20:48Z pe $

Basic XML file writer class.
It adds functionality to generateDS.generatedsuper.GeneratedsSuper.
Handles low level file I/O.

History:
V.1.0.0  2014-08-05  toh  initial version

"""

# import time
# import datetime
# import decimal
# import re
import sys
import inspect
# from nilutility.datatypes import HexInt
from six import PY2, string_types
from .base import XmlError, XmlBase
from .generateDS_ebas import ebas, gml
# from .generateDS_ebas.generateDS_ebas import \
#     Result, ResultSetType, TimeDimensionType, DataSetType, DirectPositionType, PointType, PositionType, \
#     HeightAGLType, WavelengthType, ParticleDiameterType, MeasurementUncertaintyType

class XmlWrite(XmlBase):
    """
    Partial class Xml: Part for writing Xml files from object.
    """
    def write(self, filespec=None):
        """
        Writes the file.
        Parameters:
            filespec    file name (incl. path) or file like object (stream)
                        stdout if not passed
                        
        Returns:
            None
        Raises:
            IOError           file open error
            XmlError          on any unrecoverable write error
        """
        # reset the object state for writing
        self.init_state()
        self.finalize_data()
        
        # toh: initialise XML structures
        self._set_xml()
        
        if filespec is None:
            self.file = sys.stdout
        elif isinstance(filespec, string_types):
            filename = filespec
            self.logger.info("writing file {}".format(filename))
            if PY2:
                mod = "w"
            else:
                mod = "wb"
            with open(filename, mod) as self.file:
                self._write()
        else:
            filename = None
            self.logger.debug("writing to stream")
            self.file = filespec
            self._write()

    def finalize_data(self):
        """
        Calculate data attributes that are not setable by the caller (not in 
        SETABLE_DATA_ATTRIBS). These would be redundant if set by the caller.
        Set default values for unset attributes (initialized to None).
        Parameters:
            None
        Returns:
            None
        Raises:
            various builtin exceptions when data are not set correctly.
        """
        # first set DATA if not set
        if self.data.DATA is None:
            self.data.DATA = []
            
        # calculate unsetable attributes
        self.data.NV = max(len(self.data.DATA), 0)
                
        # set default values for the rest of unset attributes
        if self.data.COMP_NAME is None:
            self.data.COMP_NAME = ['' for _unused in range(self.data.NV)]
        self._checkdims()            

    def _checkdims(self):
        """
        Checks consistency of dimensions and extents in the data structure.
        Parameters:
            None
        Returns:
            None
        Raises:
            XmlError if inconsistent
        """
        # check dimension for COMP_NAME and DATA
        if len(self.data.COMP_NAME) != self.data.NV:
            raise XmlError(
                'COMP_NAME ({} values) inconsistent with NV ({})'.format(
                                    len(self.data.COMP_NAME), self.data.NV)) 
        if len(self.data.DATA) != self.data.NV and \
           (self.data.NV != 0 or len(self.data.DATA) != 0):
            # rais error if #vars - 1 (== # dependent vars) != NV
            # special case: NV=0 and #vars = 0 (=> # dependent vars is -1!) 
            raise XmlError(
                'DATA ({} dep. variables) inconsistent with NV ({})'.format(
                                    len(self.data.DATA) - 1 , self.data.NV))
        # check variable length
        for i in range(self.data.NV-1):
            if len(self.data.DATA[0]) != len(self.data.DATA[i+1]):
                raise XmlError(
                    'number of rows for variable[0] ({}) '.format(
                                                    len(self.data.DATA[0])) +\
                    'is different from variable[{}] ({})'.format(i+1,
                                                    len(self.data.DATA[i+1])))
                
        # check variable length against time
        for i in range(self.data.NV):
            if len(self.data.SAMPLE_TIMES) != len(self.data.DATA[i]):
                raise XmlError(
                    'number of rows for SAMPLE_TIMES ({}) '.format(
                                                    len(self.data.SAMPLE_TIMES)) +\
                    'is different from variable[{}] ({})'.format(i+1,
                                                    len(self.data.DATA[i+1]))) 

    def _set_xml(self):
        """
        Sets the XML structures according to generateDS.
        Parameters:
            None
        Returns:
            None
        Raises:
            XmlError     on any unrecoverable read error
        """
        # set result set
        self.xmlResultSet = ebas.ResultSetType()
        
        # set time variable
        self.xmlTime = ebas.TimeDimensionType()
        self._set_sample_times()
        self.xmlResultSet.set_TimeDimension(self.xmlTime)
        
        # each variable will be written in one result set
        for vnum in range(self.data.NV):
            self.xmlData = ebas.DataSetType()
            for self.state in range (len(self.__class__.SECTIONS)):
                statename = self.curr_state_name()
                try:
                    method = getattr(self, '_set_' + statename.lower())
                except AttributeError:
                    # If no dedicated _set_ method exists: pass
                    pass
                else:
                    # check if method requires vnum argument
                    if 'vnum' in inspect.getargspec(method).args:
                        method(vnum)
                    else:
                        method()
            self.xmlResultSet.add_DataSet(self.xmlData)
                       
    def _set_sample_times(self):
        """
        Set the SAMPLE_TIMES XML element.
        """
        self.xmlTime.set_StartTimes(zip(*self.data.SAMPLE_TIMES)[0])
        self.xmlTime.set_EndTimes  (zip(*self.data.SAMPLE_TIMES)[1])
        
    def _set_station_code(self, vnum):
        """
        Set the STATION_CODE XML element.
        """
        self.xmlData.set_StationCode(self.data.STATION_CODE[vnum])
        
    def _set_station_name(self, vnum):
        """
        Set the STATION_NAME XML element.
        """
        self.xmlData.set_StationName(self.data.STATION_NAME[vnum])
        
    def _set_inst_type(self, vnum):
        """
        Set the INST_TYPE XML element.
        """
        self.xmlData.set_InstrumentType(self.data.INST_TYPE[vnum])
        
    def _set_comp_name(self, vnum):
        """
        Set the COMP_NAME XML element.
        """
        self.xmlData.set_ComponentName(self.data.COMP_NAME[vnum])
        
    def _set_flag_name(self, vnum):
        """
        Set the FLAG_NAME XML element. toh: not required yet
        """
        self.xmlData.set_FlagName(self.data.FLAG_NAME[vnum])
        
    def _set_matrix_name(self, vnum):
        """
        Set the MATRIX_NAME XML element.
        """
        self.xmlData.set_Matrix(self.data.MATRIX_NAME[vnum])
        
    def _set_unit(self, vnum):
        """
        Set the UNIT XML element.
        """
        self.xmlData.set_Unit(self.data.UNIT[vnum])
        
    def _set_statistic_code(self, vnum):
        """
        Set the STATISTIC_CODE XML element.
        """
        self.xmlData.set_Statistics(self.data.STATISTIC_CODE[vnum])
        
    def _set_rescode(self, vnum):
        """
        Set the RESCODE XML element.
        """
        self.xmlData.set_ResolutionCode(self.data.RESCODE[vnum])
        
    def _set_data_level(self, vnum):
        """
        Set the DATA_LEVEL XML element.
        """
        if self.data.DATA_LEVEL[vnum]:
            self.xmlData.set_DataLevel(str(self.data.DATA_LEVEL[vnum]))
        
    def _set_uncertainty(self, vnum):
        """
        Set the UNCERTAINTY XML element.
        """
        if self.data.UNCERTAINTY[vnum]:
            self.xmlData.set_MeasurementUncertainty(
            ebas.MeasurementUncertainty(unit=self.data.UNCERTAINTY[vnum][1],
                                        valueOf_=self.data.UNCERTAINTY[vnum][0]))
        
    def _set_position(self, vnum):
        """
        Set the POSITION XML element.
        """
        self.xmlData.set_Position(ebas.Position(gml.PointType(
        id="pos"+str(self.resultset_id)+'_'+str(vnum),
        srsName="urn:ogc:def:crs:EPSG:6.6:4979",
        srsDimension=len(self.data.POSITION[vnum]),
        pos=gml.DirectPositionType(
        valueOf_=self.data.POSITION[vnum]))))
        
    def _set_height_agl(self, vnum):
        """
        Set the HEIGHT_AGL XML element.
        """
        if self.data.HEIGHT_AGL[vnum]:
            self.xmlData.set_HeightAGL(
            ebas.HeightAGL(unit='m', # toh: unit for measurment height is not given
                           valueOf_=self.data.HEIGHT_AGL[vnum]))
        
    def _set_characteristics(self, vnum):
        """
        Set the XML elements given in CHARACTERISTICS.
        """
        if self.data.CHARACTERISTICS[vnum] is None or \
           len(self.data.CHARACTERISTICS[vnum]) < 1:
            return
        attr = {'Location':   {'setter': self.xmlData.set_Location,
                               'type': None},
                'Wavelength': {'setter': self.xmlData.set_Wavelength,
                               'type': ebas.WavelengthType},
                'D':          {'setter': self.xmlData.set_D,
                               'type': ebas.ParticleDiameterType},
                'Dmin':       {'setter': self.xmlData.set_Dmin,
                               'type': ebas.ParticleDiameterType},
                'Dmax':       {'setter': self.xmlData.set_Dmax,
                               'type': ebas.ParticleDiameterType},
               }
               
        for char in self.data.CHARACTERISTICS[vnum].sorted():
            if char.CT_TYPE in attr:
                inf = attr[char.CT_TYPE]
                (_skip, value, unit) = char.tuple()
                if inf['type']:
                    if unit:
                        inf['setter'](inf['type'](valueOf_=value, unit=unit))
                    else:
                        inf['setter'](inf['type'](valueOf_=value))
                else:
                    inf['setter'](value)
        
        def _set_location(self, char):
            """
            Set the Location XML element.
            """
            self.xmlData.set_Location(str(char.tuple()[1]))
            
        def _set_wavelength(self, char):
            """
            Set the Wavelength XML element.
            """
            self.xmlData.set_Wavelength(ebas.WavelengthType(
            unit=char.tuple()[2],valueOf_=char.tuple()[1]))
            
        def _set_d(self, char):
            """
            Set the D XML element.
            """
            self.xmlData.set_D(ebas.ParticleDiameterType(
            unit=char.tuple()[2],valueOf_=char.tuple()[1]))
            
        def _set_dmin(self, char):
            """
            Set the Dmin XML element.
            """
            self.xmlData.set_Dmin(ebas.ParticleDiameterType(
            unit=char.tuple()[2],valueOf_=char.tuple()[1]))
            
        def _set_dmax(self, char):
            """
            Set the Dmax XML element.
            """
            self.xmlData.set_Dmin(ebas.ParticleDiameterType(
            unit=char.tuple()[2],valueOf_=char.tuple()[1]))
        
    def _set_data(self, vnum):
        """
        Set the DATA XML element.
        """
        if len(self.data.DATA[vnum]) == 0:
            return
        self.xmlData.set_Values(self.data.DATA[vnum])
        
    def _set_flags(self, vnum):
        """
        Set the FLAGS XML element.
        """
        # toh: flags are not required yet
        self.xmlData.set_Flags(self.data.FLAGS[vnum])
        
    def _set_setkey(self, vnum):
        """
        Set the SETKEY XML element.
        """
        self.xmlData.set_setkey(self.data.SETKEY[vnum])
                
    def _write(self):
        """
        Writes the file to opened file handle.
        This method is only needed to extract the write functionality in order
        to wrap the 'with open(filename)' around it in the read() method.
        Parameters:
            None
        Returns:
            None
        Raises:
            XmlError     on any unrecoverable read error
        """
        self.xmlResultSet.export(self.file, 1, name_='ResultSet')
        self.logger.debug(u"export XML object")
           



"""
nilutility/encoding.py
$Id: encoding.py 2464 2020-07-07 14:05:22Z pe $

Module for setting the default encoding.
"""

import sys
from six.moves import reload_module

def fix_io_encoding(default='UTF-8'):
    """
    Sets the default encloding (for files, stdin/out/err if redirected etc)
    Generally:
        Is set to ascii if the environment variable PYTHONIOENCODING is not set
        else, it was set to PYTHONIOENCODING.
        For cases where PYTHONIOENCODING was not set, we can set the default
        encoding to utf-8 here.
    """
    if sys.getdefaultencoding() in (None, 'ascii'):
        # six.moves.reload_module is builtin in python2, importlib.reload in
        # python3
        reload_module(sys)
        sys.setdefaultencoding(default) # @UndefinedVariable pylint: disable=E1101
        # @UndefinedVariable: make pydev shut up
        # E1101: Module 'sys' has no 'setdefaultencoding' member



"""
nilutility/timer.py
$Id: timer.py 1374 2016-09-07 08:58:19Z pe $

Timer class (measure elapsed walltime)
"""

import time

class Timer(object):
    """
    Wall clock timer class.
    """
    def __init__(self):
        """
        Object initialization. Starts timer automatically.
        However, the timer can be started explicitely later by calling start().
        """
        self.starttime = None
        self.stoptime = None
        self.start()

    def start(self):
        """
        Start the timer.
        Parameters:
            None
        Returns:
            None
        """
        self.starttime = time.time()

    def stop(self):
        """
        Stop the timer.
        Parameters:
            None
        Returns:
            None
        """
        self.stoptime = time.time()
    
    def elapsed(self):
        """
        Get timer wall clock time elapsed.
        Parameters:
            None
        Returns:
            time elapsed in seconds (float)
        """
        return self.stoptime - self.starttime



"""
$Id: fl.py 2427 2020-03-19 23:09:09Z pe $

EBAS Masterdata Class for Flags

This module implements the class EbasMasterFL.

History:
V.1.0.0  2014-02-25  pe  initial version

"""

from .offline_masterdata import FLOfflineMasterData
from .base import EbasMasterBase

class EbasMasterFL(EbasMasterBase, FLOfflineMasterData):
    """
    Domain Class for flags.
    Objects of this class do not represent entities, this class provides class
    methods for flag handling.
    """

    # Problem: for checking validity of flags and consistency of flags vs.
    # values values without DB connection, we need an offline store of flag
    # masterdata.
    # read offline masterdata for FL (flags)
    # Those are fallback values, will be read from database if possible.
    FLOfflineMasterData.read_pickle_file()

    def __init__(self, dbh=None):
        """
        Read cached fl masterdata if dbh is provided.
        """
        FLOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)



"""
$Id: zt.py 2088 2018-06-29 13:46:38Z pe $

EBAS Masterdata Class for zero/spancheck type masterdata

This module implements the class EbasMasterZT.

History:
V.1.1.0  2016-10-17  pe  initial version

"""

from .offline_masterdata import ZTOfflineMasterData
from .base import EbasMasterBase

class EbasMasterZT(EbasMasterBase, ZTOfflineMasterData):
    """
    Domain Class for zero/spancheck type masterdata.
    Objects of this class do not represent entities, this class provides class
    methods for handling stations and checking them against master data.
    Stations master data are retrieved from database or from offline storage
    when no database access is possible.
    """

    # read offline masterdata for ZT (zero/spancheck type)
    # Those are fallback values, will be read from database as soon as possible.
    ZTOfflineMasterData.read_pickle_file()

    def __init__(self, dbh=None):
        """
        Read zt masterdata if dbh is provided.
        """
        ZTOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)



from .generated.ebas_generated import *
from .generated.ebas_generated import GeneratedsSuper, showIndent, __all__

# "hook" gml Point into the ebas scheme
def _PosExportChildren(self, outfile, level, namespace_='ebas:', name_='PositionType', fromsubclass_=False, pretty_print=True):
    if pretty_print:
        eol_ = '\n'
    else:
        eol_ = ''
    if self.Point is not None:
        # pe:
        # < showIndent(outfile, level, pretty_print)
        # < outfile.write('<%sPoint>%s</%sPoint>%s' % (namespace_, self.gds_format_string(quote_xml(self.Point).encode(ExternalEncoding), input_name='Point'), namespace_, eol_))
        self.Point.export(outfile, level, namespace_='gml:', name_='Point', pretty_print=pretty_print)
Position.exportChildren = _PosExportChildren


# Add "Z" to datetime format
GeneratedsSuper.orig_gds_format_datetime = GeneratedsSuper.gds_format_datetime
def _gds_format_datetime(self, input_data, input_name=''):
    return self.orig_gds_format_datetime(input_data, input_name) + 'Z'
GeneratedsSuper.gds_format_datetime = _gds_format_datetime

# make Start and End Times lists of datetime
# This could partly also be done by changing GeneratedsSuper, but this way all
# changes are coded in this file.
def _gds_format_datetime_list(self, input_data, input_name=''):
    res = []
    for dat in input_data:
        res.append(self.gds_format_datetime(dat, input_name))
    return ' '.join(res)
GeneratedsSuper.gds_format_datetime_list = _gds_format_datetime_list
def _TdtExportChildren(self, outfile, level, namespace_='ebas:', name_='TimeDimensionType', fromsubclass_=False, pretty_print=True):
    if pretty_print:
        eol_ = '\n'
    else:
        eol_ = ''
    if self.StartTimes is not None:
        showIndent(outfile, level, pretty_print)
        outfile.write('<%sStartTimes>%s</%sStartTimes>%s' % (namespace_, self.gds_format_datetime_list(self.StartTimes, input_name='StartTimes'), namespace_, eol_))
    if self.EndTimes is not None:
        showIndent(outfile, level, pretty_print)
        outfile.write('<%sEndTimes>%s</%sEndTimes>%s' % (namespace_, self.gds_format_datetime_list(self.EndTimes, input_name='EndTimes'), namespace_, eol_))
TimeDimensionType.exportChildren = _TdtExportChildren

# correct version of format_float and gds_format_float_list
# This could also be done by changing GeneratedsSuper, but this way all changes
# are coded in this file.
def _gds_format_float(self, input_data, input_name=''):
    if input_data is None:
        return 'NaN'
    return str(input_data)
def _gds_format_float_list(self, input_data, input_name=''):
    res = []
    for val in input_data:
        res.append(self.gds_format_float(val, input_name))
    return ' '.join(res)
GeneratedsSuper.gds_format_float = _gds_format_float
GeneratedsSuper.gds_format_float_list = _gds_format_float_list



"""
$Id: file.py 2494 2020-09-16 12:31:19Z pe $

EBAS I/O file
"""

from .check_metadata import EbasFileCheckMetadata
from .check_data import EbasFileCheckData
from .write import EbasFileWrite
from nilutility.datatypes import DataObject
try:
    # pylint: disable=W0611
    # W0611: Unused import
    # used by client classes
    from ebas.domain.entities.interface_ebas_io.io_from_domain \
        import EbasFileFromDomain  #@UnusedImport
except ImportError:
    from .from_domain import EbasFileFromDomain
from .read import EbasFileRead


class EbasFile(EbasFileCheckMetadata, EbasFileCheckData, EbasFileFromDomain,
               EbasFileRead, EbasFileWrite):
    # pylint: disable=W0223
    # W0223: ... is abstract in base class but is not overridden
    # this class is abstract as well (does not have any NotImplementedError)
    """
    This is a base class for all types of Ebas IO files (nasa_ames, xml, csv).
    Its an abstract, partial class.
    """
    pass

class EbasJoinedFile(EbasFile):
    """
    This is a base class for Ebas IO files which can have multiple metadata
    intervals (e.g. OPeNDAP, NetCDF).
    The join method for those objects will generate an new object of the same
    type with aggregated key metadata and the original file objects in the
    _metadata_intervals attribute. This attribute will be transparently
    accessible as a property method.
    """
    def __init__(self, metadata_intervals=None, *args, **kwargs):
        """
        Class initialization.
        """
        EbasFile.__init__(self, *args, **kwargs)
        self._metadata_intervals = metadata_intervals
        self._prepare_write = False  # avoid multiple cascading down to kids

    @property
    def metadata_intervals(self):
        """
        Property for metadata intervals.
        List of lists ([[start, end, metadata],...]
        This method makes using the original file object or a joined file
        object (with multiple intervals) completely transparent.
        """
        if self._metadata_intervals is None:
            if not self.sample_times:
                self._metadata_intervals = []
            else:
                self._metadata_intervals = [
                    [self.sample_times[0][0], self.sample_times[-1][1], self]]
        return self._metadata_intervals

    def get_metadata_set(self, element, string=False, exclude_aux=False):
        """
        Generates a set of different values for one metadata element (value of
        global metadata and all values per variable).
        If emement is a tuple, a list of combinations of metadata elements
        is returned.
        Overriden in order to take care of multiple metadata intervals.
        This method is meant to give "all different" values for a metadata
        element.
        Parameter:
            element     metadata element (string or dict)
            string      whether the elements should be reresented as str
            exclude_aux
                        do not report meatdata for auxiliary variables
        Returns:
            list (list with unique elements, i.e. set)
        """
        # Because metadata values can be unhashable objects (dict, DataObject),
        # this method must be implemented as a list, not as a set.

        def _add(x):
            if x not in set_: set_.append(x)

        # do the base method (look in my own metadata, if set
        set_ = super(EbasJoinedFile, self).get_metadata_set(
            element, string=string, exclude_aux=exclude_aux)

        # look recursivly in all metadata intervals (which do not refer to self)
        for fil in [elem[2] for elem in self.metadata_intervals
                    if elem[2] is not self]:
            for elem in fil.get_metadata_set(element, string, exclude_aux):
                _add(elem)
        return set_

    def prepare_write(self, open_file, *args, **kwargs):
        """
        Prepare the IO object for writing.
        Parameters:
            open_file  wheter the oputut file should be opened and the file
                       handle be returned. (Should net be done for certain
                       types of output objects (e.g. EbasNetCDF) or on the
                       atomic files within an EbasJoinedFile container)
            datadef    data definition (EBAS_1 or EBAS_1.1)
        Returns
            None
        """
        super(EbasJoinedFile, self).prepare_write(open_file, *args,
                                                  **kwargs)
        self._prepare_write = True  # avoid multiple cascading to kids
        for obj in [x[2] for x in self.metadata_intervals if x[2] != self]:
            # call intervall object's method (but without opening the files!
            obj.prepare_write(False, *args, **kwargs)
        self._prepare_write = False

    def setup_ebasmetadata(self, *args, **kwargs):
        # call base class method for self:
        super(EbasJoinedFile, self).setup_ebasmetadata(*args, **kwargs)
        if self._prepare_write:
            return  # avoid multiple cascading (prepare_write cascades already)
        for obj in [x[2] for x in self.metadata_intervals if x[2] != self]:
            # call intervall object's method
            obj.setup_ebasmetadata(*args, **kwargs)
        
    def set_timespec_metadata(self, *args, **kwargs):
        # call base class method for self:
        super(EbasJoinedFile, self).set_timespec_metadata(*args, **kwargs)
        if self._prepare_write:
            return  # avoid multiple cascading (prepare_write cascades already)
        for obj in [x[2] for x in self.metadata_intervals if x[2] != self]:
            # call intervall object's method
            obj.set_timespec_metadata(*args, **kwargs)

    def set_default_metadata(self, *args, **kwargs):
        # call base class method for self:
        super(EbasJoinedFile, self).set_default_metadata(*args, **kwargs)
        if self._prepare_write:
            return  # avoid multiple cascading (prepare_write cascades already)
        for obj in [x[2] for x in self.metadata_intervals if x[2] != self]:
            # call intervall object's method
            obj.set_default_metadata(*args, **kwargs)

    def consolidate_metadata(self, *args, **kwargs):
        # call base class method for self:
        super(EbasJoinedFile, self).consolidate_metadata(*args, **kwargs)
        if self._prepare_write:
            return  # avoid multiple cascading (prepare_write cascades already)
        for obj in [x[2] for x in self.metadata_intervals if x[2] != self]:
            # call intervall object's method
            obj.consolidate_metadata(*args, **kwargs)

    def union_projects(self):
        """
        Get a union of all projects in the file.
        Base implementation in EbasFileBase, but here overridden for adequate
        handling of joined files.
        Parameters:
            None
        Returns:
            list of project acronyms
        """
        if len(self.metadata_intervals) == 1 and \
                self.metadata_intervals[0][2] is self:
            return super(EbasJoinedFile, self).union_projects()
        return sorted(
            {prj
             for x in self.metadata_intervals
             for prj in x[2].union_projects()})

    def common_glob_metadata(self, other):
        """
        Calculate the common metadata of two objects (or variables of two
        objects).
        Parameters:
            other     the other object (for comparison of the metadata)
        Returns:
            DataObject containing the common metadata
        """
        # merge global metadata, generally intersection of both files
        new_metadata = DataObject()
        for key in set(list(self.metadata.keys()) + \
                       list(other.metadata.keys())):
            if self.metadata[key] == other.metadata[key]:
                new_metadata[key] = self.metadata[key]
            elif key in ('submitter', 'originator'):
                # submitter, originator: union of both files:
                # (don't use set() here, we want to keep the order as good as
                # possible
                new_metadata[key] = self.metadata[key] + \
                                    [x for x in other.metadata[key]
                                     if x not in self.metadata[key]]
            else:
                new_metadata[key] = None
        return new_metadata


    def common_var_metadata(self, other, vnum, glob_metadata):
        """
        Calculate the common metadata of two objects (or variables of two
        objects).
        Parameters:
            other     the other object (for comparison of the metadata)
            vnum      the variable numbers in the file which should be compared
            glob_metadata
                      global metadata for the new, joined file (only needed
        Returns:
            DataObject containing the common variable metadata
        """
        res = DataObject()
        # iterate over all keys in both file's global and per variable metadata:
        for key in set(list(self.metadata.keys()) + \
                       list(other.metadata.keys()) + \
                       list(self.variables[vnum].metadata.keys()) + \
                       list(other.variables[vnum].metadata.keys())):
            if self.get_meta_for_var(vnum, key) == \
               other.get_meta_for_var(vnum, key) and \
               self.get_meta_for_var(vnum, key) != glob_metadata[key]:
                # only set variable metadata when not equal global metadata
                # otherwise problem: e.g. regime code is None in this stage
                # (only set in set_default later), thus var metadata would be
                # set None here. Later default would be set to IMG in global
                # metadata, but var metadata would still be None -> wrong.
                res[key] = self.get_meta_for_var(vnum, key)
            elif key in ['submitter', 'originator']:
                # submitter, originator: union of both files:
                # (don't use set() here, we want to keep the order as good as
                # possible
                val = self.get_meta_for_var(vnum, key) + \
                      [x for x in other.get_meta_for_var(vnum, key)
                       if x not in self.get_meta_for_var(vnum, key)]
                if val != glob_metadata[key]:
                    res[key] = val
            elif key == 'revdate':
                # special case for revdate: use maximum
                val1 = self.get_meta_for_var(vnum, key)
                val2 = other.get_meta_for_var(vnum, key)
                if val1 and not val2 and val1 != glob_metadata[key]:
                    res[key] = val1
                elif val2 and not val1 and val2 != glob_metadata[key]:
                    res[key] = val2
                elif val1 and val2:
                    val = max(val1, val2)
                    if val != glob_metadata[key]:
                        res[key] = val
                # else: not set for variable
            # else: not set for variable
        return res

    def join(self, other):
        """
        Join self with other if possible.
        Default implementation in EbasFile: do not join. This is overridden for
        EbasJoinedFile.
        Parameters:
            other    other EbasIO object
        Returns
            New, joined EbasIO object
            None (no join possible)
        """
        # check if objects can be joined:
        # same variables, same setkeys, data interval can be joined
        if len(self.variables) != len(other.variables) or \
           self.get_metadata_list_per_var('setkey') != \
               other.get_metadata_list_per_var('setkey'):
            return None
        # make a new object of the same class then self:
        new = self.__class__(
            indexdb=self.indexdb, metadata_intervals=sorted(
                self.metadata_intervals + other.metadata_intervals))
        new.from_domain = self.from_domain
        new.metadata = self.common_glob_metadata(other)
        new.sample_times = self.sample_times + other.sample_times
        new.variables = []
        for vnum in range(len(self.variables)):
            var1 = self.variables[vnum]
            var2 = other.variables[vnum]
            metadata = self.common_var_metadata(other, vnum, new.metadata)
            new.variables.append(
                DataObject(
                    is_hex=var1.is_hex,
                    values_=var1.values_ + var2.values_,
                    flags=var1.flags + var2.flags,
                    flagcol=True,
                    metadata=metadata,
                    ))            
            # characteristics will be copied as "common_metadata" above
            # because same setkey will produce same characteristics for all
            # intervals
        return new

    def _var_fits_in_file(self, di_obj, time):  # pylint: disable-msg=R0911
        # R0911: Too many return statements
        #  --> clear to read (better then more complex if clauses
        """
        Check whether the variable fits in this file (either same submission
        or all metadata and the timesseries must be matching).
        This is a less strict implementation for EbasJoinedFile which overrides
        the EbasFileFromDomain._var_fits_in_file implementation.
        Parameters:
            di_obj      di object to be processed
            time        time inteval to be processed (start, end)
        Returns:
            True/False (fits / does not)
            overlap (DatetimeInterval) if it partly fits
        """
        return self._var_fits_in_file_general(di_obj, time)



"""
EANET filter file read
"""

import os
import re
import datetime
from collections import defaultdict
from fileformats.CsvUnicode import CsvUnicodeReader
from nilutility.datetime_helper import DatetimeInterval
from .base import EanetError
from .base import EanetFilterBase, EanetFilterVariable


class EanetFilterRead(EanetFilterBase):
    """
    Class for reading EANET filter files
    """

    def __init__(self):
        """
        Initialize class.
        """
        EanetFilterBase.__init__(self)
        self._rawdata = None
        self._sample_time_flags = []

    def read(self, filename):
        """
        Read from file.
        Parameters:
            filename    path and file name
        Returns:
            None
        Raises:
            IOError     in case of I/O issues
            EanetError  other unrecoverable raeding errors
        """
        self.logger.info('reading file %s', filename)
        self.file_name = filename
        self.parse_filename()
        self._read_file()
        if self.errors:
            raise EanetError("Exiting because of previous errors")

    def parse_filename(self):
        """
        Parses site code from filename, warnings if filename is not standard.
        Parameters:
            None
        Returns:
            None
        """
        basename = os.path.basename(self.file_name)
        reg = re.match(
            '([A-Z]{3}[0-9]{3})_([0-9]{4})_(dry_deposition_filter_pack).(csv)',
            basename)
        if not reg:
            self.logger.warning("Unconventional file name '%s'.", basename)
            reg = re.match('([A-Z]{3}[0-9]{3})_', basename)
            if not reg:
                self.logger.warning(
                    "Unable to parse site code from file name.")
                return
        self.site_code = reg.group(1)

    def _read_file(self):
        """
        Read the file content.
        Parameters:
            None
        Returns:
            None
        Raises: EanetError in case of unrecoverable errors
        """
        with open(self.file_name, 'rb') as fil:
            rdr = CsvUnicodeReader(fil)
            self._rawdata = [x for x in rdr]
        self._check_headerline()
        self._check_units()
        self._check_column_numbers()
        self._read_country()
        self._read_site_name()
        self._read_sample_times()
        self._read_variables()
        self.notes = [None if x[-1] == '' else x[-1] for x in self._rawdata[2:]]

    def _check_headerline(self):
        """
        Check the header line (line2).
        """
        if self._rawdata[1][:6] != ['Country', 'Site', 'Date', 'Time', 'Date',
                                    'Time']:
            self.error(
                'Format error: The first 6 elements in the second line must be '
                "'Country', 'Site', 'Date', 'Time', 'Date', 'Time'",
                row=1, col=0)
        if self._rawdata[1][-1] == 'Note':
            # Notes in the last column, read last column as notes, not as var
            self._hasnotes = True
            self._lastcol = len(self._rawdata[1]) - 1
        else:
            # No Notes in last column, read all columns as data
            self._hasnotes = False
            self._lastcol = len(self._rawdata[1])
        for i, varname in enumerate(self._rawdata[1][6:self._lastcol]):
            if varname.strip() == '':
                self.error("No variable name", row=1, col=i)
                # in order to be able to recover from this error:
                self._rawdata[1][i] = 'NoName{}'.format(i-5)

    def _check_units(self):
        """
        Checks line 1 of the file: Units
        """
        if any([self._rawdata[0][i] != '' for i in range(5)]):
            self.error(
                'Format error: The first 5 elements in the first line must be '
                'empty.', row=0, col=0)
        if self._rawdata[0][5] != u'Unit':
            self.error(
                "Format error: The 6. element in the first line must be 'Unit'",
                row=0, col=5)
        if self._hasnotes and self._rawdata[0][-1] != '':
            self.error(
                "Format error: The last element in the first line "
                "(unit for 'Note') must be empty",
                row=0, col=len(self._rawdata[0]-1))
        for i, unit in enumerate(self._rawdata[0][6:self._lastcol]):
            if unit.strip() == '':
                self.error("No unit", row=0, col=i)

    def _check_column_numbers(self):
        """
        Check the number of columns in all lines is equal.
        Parameters:
            None
        Returns:
            None
        Raises: EanetError in case of unrecoverable errors
        """
        occ = defaultdict(list)  # occurrences of different column counts
        for i, row in enumerate(self._rawdata):
            occ[len(row)].append(str(i+1))
        if len(occ) > 1:
            self.error("Different column numbers in file: " + \
                       self._occ_to_str(occ, add_spec='columns'))
            # unrecoverable error
            raise EanetError("Exiting because of previous errors")

    def _read_country(self):
        """
        Read the country column.
        """
        occ = defaultdict(list)  # occurrences of different countries
        for i, row in enumerate(self._rawdata[2:]):
            occ[row[0]].append(str(i+3))
        if len(occ) > 1:
            self.error("Different country names in file (column 1): " +\
                       self._occ_to_str(occ))
        self.country_name = self._rawdata[2][0]

    def _read_site_name(self):
        """
        Read the site name column.
        """
        occ = defaultdict(list)  # occurrences of different sites
        for i, row in enumerate(self._rawdata[2:]):
            occ[row[1]].append(str(i+3))
        if len(occ) > 1:
            self.error("Different site names in file (column 2): " +\
                       self._occ_to_str(occ))
        self.site_name = self._rawdata[2][1]

    def _read_sample_times(self):
        """
        Read the sample times from the file's rawdata array.
        Parameters:
            None
        Returns:
            None
        """
        self.sample_times = []
        self._sample_time_flags = []
        for i, row in enumerate(self._rawdata[2:]):
            flags = []
            start = self._parse_datetime(i+2, 2)
            end = self._parse_datetime(i+2, 4)
            try:
                sample = DatetimeInterval(start, end)
                # DatetimeInterval checks for start < end
            except (TypeError, ValueError) as excpt:
                self.error(
                    "Start/end date/time: %s", str(excpt),
                    row=i+2, col=2)
                sample = DatetimeInterval(None, None)
            if i and self.sample_times[i-1][1] and \
                    sample[0] and self.sample_times[i-1][1] > sample[0]:
                # overlap!
                if sample[0].second == 0 and sample[0].minute == 0 and \
                        sample[0].hour == 0:
                    self.warning(
                        "sample time overlap %s %s. Changing start time to %s "
                        "and flagging with flag 249.",
                        self.sample_times[i-1], sample,
                        self.sample_times[i-1][1],
                        row=i+2, col=2)
                    sample[0] = self.sample_times[i-1][1]
                    flags.append(249)
                else:
                    self.warning(
                        "sample time overlap %s %s. Changing end time to %s and "
                        "flagging with flag 249.",
                        self.sample_times[i-1], sample, sample[0],
                        row=i+1, col=4)
                    self.sample_times[i-1][1] = sample[0]
                    self._sample_time_flags[-1].append(249) 
            self.sample_times.append(sample)
            self._sample_time_flags.append([])

    def _read_variables(self):
        """
        Read all data from the file's rawdata array.
        Parameters:
            None
        Returns:
            None
        """
        for i in range(6, self._lastcol):
            values, flags = self._get_values(i)
            self.variables.append(EanetFilterVariable(
                self._rawdata[1][i].strip(), self._rawdata[0][i].strip(),
                values, flags))

    def _get_values(self, i):
        """
        Get the values for one variable.
        Parameters:
            i      column index
        Returns:
            list of values (float, None for missing)
        """
        def _cvt(str_val, var_name, i, j):
            """
            Convert value to float or None.
            Parameters:
                str_val  string value
            Returns:
                float, None for missing
            """
            if str_val == '':
                return None, [999]
            flags = []
            if str_val[0] == '<':
                str_val = str_val[1:].strip()
                flags = [781]
            try:
                return float(str_val), flags
            except ValueError as excpt:
                self.error(
                    "Syntax error in value for variable %s: %s",
                    var_name, str(excpt),
                    row=j, col=i)
                return None, [999]

        values = []
        flags = []
        for j, val in enumerate(self._rawdata[2:], 2):
            value, flag = _cvt(val[i], self._rawdata[1][i], i, j)
            flag += self._sample_time_flags[j-2]
            values.append(value)
            flags.append(flag)
        return values, flags

    @staticmethod
    def _occ_to_str(occ, add_spec=''):
        """
        Generates string for error messages in case of different occurences.
        Parameters:
            occ        occurence dictionary (occurences: list(line numbers))
            add_spec   additional specifier (optional)
        Returns:
            str summarising the different occurrences (used for error messages)
        """
        if add_spec:
            add_spec = ' ' + add_spec
        return ", ".join(
            ["{} {} containing {}{} ({})".format(
                len(x[1]), 'rows' if len(x[1]) > 1 else 'row',
                x[0],
                add_spec,
                ('lines ' if len(x[1]) > 1 else 'line ') +
                ', '.join(
                    (x[1][:3] + ['...']) if len(x[1]) > 3 else x[1]))
             for x in sorted(
                 occ.iteritems(),
                 key=lambda k, v: len(v),
                 reverse=True)])



"""
Masterdata caches (submodule masterdata):

A module to cache all important masterdata in EBAS. Those masterdata are mainly
lists of controlled vocabulary (e.g. parameters, instrument types, ...).

Those parts of the code which do not have direct access to the database
(e.g. ebas-io module running on a user's machine) use this caches. Parts of
code which have database access, update those caches right after connecting.

The storage technology is simple python pickle files.
"""
import logging
from .am import EbasMasterAM
from .ac import EbasMasterAC
from .ax import EbasMasterAX
from .ay import EbasMasterAY
from .bc import EbasMasterBC
from .ca import EbasMasterCA
from .cc import EbasMasterCC
from .co import EbasMasterCO
from .cy import EbasMasterCY
from .dc import DCBase
from .dl import EbasMasterDL
from .fl import EbasMasterFL
from .fm import EbasMasterFM
from .ft import EbasMasterFT
from .ht import EbasMasterHT
from .im import EbasMasterIM
from .ip import EbasMasterIP
from .it import EbasMasterIT
from .ma import EbasMasterMA
from .md import EbasMasterMD
from .oc import EbasMasterOC
from .org import EbasMasterOR
from .pg_pl import EbasMasterPG, EbasMasterPL
from .pm import EbasMasterPM
from .pr import EbasMasterPR
from .ps import EbasMasterPS
from .qm_qv import EbasMasterQM, EbasMasterQV
from .re import EbasMasterRE
from .sc import EbasMasterSC
from .se import EbasMasterSE
from .sm import EbasMasterSM
from .sp import EbasMasterSP
from .st import EbasMasterST
from .sw import EbasMasterSW
from .sx import EbasMasterSX
from .sz import EbasMasterSZ
from .vp import EbasMasterVP
from .vt import EbasMasterVT
from .wc import EbasMasterWC
from .zn import EbasMasterZN
from .zt import EbasMasterZT


def read_all_caches_from_db(dbh):
    """
    Reads all chache masterdata from db. This can be called by programs that
    have db access and want to make sure all called modules work with current
    master data from the database.
    Parameters:
        dbh    database handle
    Returns:
        None
    """
    EbasMasterAC(dbh)
    EbasMasterAM(dbh)
    EbasMasterAX(dbh)
    EbasMasterAY(dbh)
    EbasMasterBC(dbh)
    EbasMasterCA(dbh)
    EbasMasterCC(dbh)
    EbasMasterCO(dbh)
    EbasMasterCY(dbh)
    DCBase(dbh)
    EbasMasterDL(dbh)
    EbasMasterFL(dbh)
    EbasMasterFM(dbh)
    EbasMasterFT(dbh)
    EbasMasterHT(dbh)
    EbasMasterIM(dbh)
    EbasMasterIP(dbh)
    EbasMasterIT(dbh)
    EbasMasterMA(dbh)
    EbasMasterMD(dbh)
    EbasMasterOC(dbh)
    EbasMasterOR(dbh)
    EbasMasterPG(dbh)
    EbasMasterPL(dbh)
    EbasMasterPM(dbh)
    EbasMasterPR(dbh)
    EbasMasterPS(dbh)
    EbasMasterQM(dbh)
    EbasMasterQV(dbh)
    EbasMasterRE(dbh)
    EbasMasterSC(dbh)
    EbasMasterSE(dbh)
    EbasMasterSM(dbh)
    EbasMasterSP(dbh)
    EbasMasterST(dbh)
    EbasMasterSW(dbh)
    EbasMasterSX(dbh)
    EbasMasterSZ(dbh)
    EbasMasterVP(dbh)
    EbasMasterVT(dbh)
    EbasMasterWC(dbh)
    EbasMasterZN(dbh)
    EbasMasterZT(dbh)



"""
$Id: offline_masterdata.py 2743 2021-11-23 15:48:15Z pe $

EBAS Offline Masterdata classes
"""

import os
from ebas.domain.masterdata.offline_masterdata.base import OfflineMasterDataBase


class ACOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for AC
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'AC.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_ac()
        cls.INIT = True

class AMOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for AM
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'AM.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_am()
        cls.INIT = True

class AXOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for AX
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'AX.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_ax()
        cls.INIT = True


class AYOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for AY
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'AY.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_ay()
        cls.INIT = True


class BCOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for BC and BA
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'BC.pkl')
    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_bc()
        # set all characteristics to []:
        for bc_id in cls.META:
            cls.META[bc_id]['characteristics'] = {}
        for ba_ in dbh.getcache_ba():
            if ba_['CT_DATATYPE'] == 'CHR':
                val = ba_['BA_VAL_CHR']
            elif ba_['CT_DATATYPE'] == 'INT':
                val = ba_['BA_VAL_INT']
            else:
                val = ba_['BA_VAL_DBL']
            cls.META[ba_['BC_ID']]['characteristics'][ba_['CT_TYPE']] = val
        cls.INIT = True


class CAOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for CA
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'CA.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_ca()
        cls.INIT = True


class CCOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for CC
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'CC.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_cc()
        cls.INIT = True


class COOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for CO
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'CO.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_co()
        cls.INIT = True


class CXOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for CX
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'CX.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_cx()
        cls.INIT = True


class CYOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for CY
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'CY.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_cy()
        cls.INIT = True


class CTOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for CT and CV
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'CT_CV.pkl')
    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handke
        Returns:
            None
        """
        cls.META['SORTORDER'] = list(dbh.get_ct_order())
        cls.META['CHARACTERISTICS'] = {}
        for ct_ in cls.META['SORTORDER']:
            cls.META['CHARACTERISTICS'][ct_] = dbh.select_ct(ct_)
        cls.META['VALIDITY'] = list(dbh.selectlist_cv())
        cls.INIT = True


class DLOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for DL
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'DL.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_dl()
        cls.INIT = True


class FLOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for FL
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'FL.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_fl()
        cls.INIT = True


class FMOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for FM
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'FM.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_fm()
        cls.INIT = True


class FPOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for FP
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'FP.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_fp()
        cls.INIT = True

class FTOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for FT
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'FT.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_ft()
        cls.INIT = True

class IMOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for IM
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'IM.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_im()
        cls.INIT = True


class HTOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for HT
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'HT.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_ht()
        cls.INIT = True


class IPOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for IP.
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'IP.pkl')
    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_ip()
        cls.INIT = True


class ITOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for IT.
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'IT.pkl')
    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_it()
        cls.INIT = True


class MAOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for MA
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'MA.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_ma()
        cls.INIT = True


class MDOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for MD
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'MD.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_md()
        cls.INIT = True


class OCOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for OC
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'OC.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_oc()
        cls.INIT = True


class OROfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for OR
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'OR.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_or()
        cls.INIT = True


class PMOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for PM.
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'PM.pkl')
    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_pm()
        cls.INIT = True


class PGOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for PG.
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'PG.pkl')
    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_pg()
        cls.INIT = True


class PLOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for PL.
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'PL.pkl')
    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_pl()
        cls.INIT = True


class PROfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for PR
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'PR.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_pr()
        cls.INIT = True


class PSOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for PS
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'PS.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_ps()
        cls.INIT = True


class QMOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for QM
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'QM.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_qm()
        cls.INIT = True


class QOOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for QO
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'QO.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_qo()
        cls.INIT = True


class QVOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for QV
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'QV.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_qv()
        cls.INIT = True


class REOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for RE
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'RE.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_re()
        cls.INIT = True


class SCOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for SC
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'SC.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_sc()
        cls.INIT = True


class SEOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for SE
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'SE.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_se()
        cls.INIT = True


class SMOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for SM
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'SM.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_sm()
        cls.INIT = True


class SPOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for SP
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'SP.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_sp()
        cls.INIT = True


class STOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for ST
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'ST.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_st()
        cls.INIT = True


class SXOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for SX
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'SX.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_sx()
        cls.INIT = True


class SZOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for SZ
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'SZ.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_sz()
        cls.INIT = True


class SWOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for SW
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'SW.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_sw()
        cls.INIT = True


class VPOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for VP
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'VP.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_vp()
        cls.INIT = True


class VTOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for VT
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'VT.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_vt()
        cls.INIT = True


class WCOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for WC
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'WC.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_wc()
        cls.INIT = True


class ZNOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for ZN
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'ZN.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_zn()
        cls.INIT = True


class ZTOfflineMasterData(OfflineMasterDataBase):
    """
    Base class for handling Offline Masterdata for ZT
    """
    # file name
    META_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'ZT.pkl')

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.META = dbh.getcache_zt()
        cls.INIT = True
